import os
import json
from ovo.app.components.molstar_custom_component.dataclasses import StructureVisualization, ChainVisualization
from ovo.app.components.molstar_custom_component.contigs_parser import ContigsParser
import uuid
import glob

parent_dir = os.path.dirname(os.path.abspath(__file__))
build_dir = os.path.join(parent_dir, "frontend/build")

_component_func = None


def molstar_custom_component(
    structures: list[StructureVisualization],
    key: str | None,
    height: str | int = "500px",
    width: str | int = "100%",
    show_controls=False,
    selection_mode=False,
    download_filename=None,
    html_filename=None,
    force_reload=False,
):
    """Create a new instance of "molstar_custom_component".

    Parameters
    ----------
    structures: list[StructureVisualization]
        A list containing definitions for the structure visualizations.
    height: str
        Height of the component
    width: str
        Width of the component
    show_controls: bool
        Show panels with tools and sequence
    selection_mode: bool
        Turn on selection mode
    download_filename : str or None, optional
        If provided, enables a download button for the structure's PDB files with this filename.
    html_filename : str or None, optional
        If provided, enables a button to download the current visualization as an HTML file with this filename.
    key: str or None
        An optional key that uniquely identifies this component. If this is
        None, and the component's arguments are changed, the component will
        be re-mounted in the Streamlit frontend and lose its current state.
    force_reload: bool, default False
        although component should reload all settings automatically, sometimes it's useful to force reload of the component,
        should be turned on only when something has just changed (such as highlighted selections)

    Returns
    -------
    Data from Mol* in JSON format.

    """
    global _component_func
    if _component_func is None:
        import streamlit.components.v1 as components

        _component_func = components.declare_component("molstar_custom_component", path=build_dir)

    serializedStructures = json.dumps([struct.to_dict() for struct in structures])

    component_value = _component_func(
        structures=serializedStructures,
        height=f"{height}px" if isinstance(height, int) else height,
        width=f"{width}px" if isinstance(width, int) else width,
        showControls=show_controls,
        selectionMode=selection_mode,
        forceReload=force_reload,
        key=key,
    )

    if download_filename:
        import streamlit as st

        for structure in structures:
            st.download_button(
                label="Download PDB",
                data=structure.pdb,
                file_name=f"{download_filename}.pdb",
            )
            break

    if html_filename:
        import streamlit as st

        st.download_button(
            "Download HTML",
            data=molstar_html(structures),
            file_name=f"{html_filename}.html",
            mime="text/html",
        )

    return component_value


_static_files = None


def read_static_files():
    global _static_files
    if _static_files is None:
        css_files = sorted(glob.glob(os.path.join(build_dir, "chunk-*.css")))
        if not css_files:
            raise FileNotFoundError("No CSS file found in the build directory.")
        with open(css_files[0]) as f:
            css_content = f.read()
        js_files = sorted(glob.glob(os.path.join(build_dir, "chunk-*.js")))
        if not js_files:
            raise FileNotFoundError("No JS file found in the build directory.")
        with open(js_files[0]) as f:
            js_content = f.read()
        _static_files = css_content, js_content
    return _static_files


def molstar_html(structures: list[StructureVisualization]):
    css_content, js_content = read_static_files()

    return f"""
            <div id="root" class="molstar_notebook"></div>
            <style type="text/css">
            {css_content}
            </style>
            <script>
            window.STRUCTURES = {json.dumps([s.to_dict() for s in structures])};
            
            // JS bundle
            {js_content.replace('</script>"', '</" + "script>"')}
            // watch out that here the original js_content might also include closing script tags NOT in quotes!
            // the problem is that doing things like replacing with </ script> will break the code functionality anyway...
            </script>
            """


def molstar_notebook(
    structures: list[StructureVisualization],
    height="500px",
    width="800px",
):
    from IPython.display import display, HTML, Javascript

    html_data = molstar_html(structures).strip()
    assert html_data.endswith("</script>"), "Molstar HTML should end with </script>"

    wrapper_id = f"molstar_{uuid.uuid4()}"

    # JavaScript code to create a Blob URL and a download link
    # We turn "...</script>" into "...</script" + ">" to avoid closing script tag in VScode
    js_code = f"""
    setTimeout(function(){{
        var wrapper = document.getElementById("{wrapper_id}")
        if (wrapper === null) {{
            throw new Error("Wrapper element #{wrapper_id} not found anymore")      
        }}
        var blob = new Blob([{json.dumps(html_data[:-1])} + ">"], {{ type: 'text/html' }});
        var url = URL.createObjectURL(blob);
    
        // Create the iframe
        var iframe = document.createElement('iframe');
        iframe.src = url;
        iframe.style = "border: 0; width: {width}; height: {height}"
        iframe.allowFullscreen = true;
        wrapper.appendChild(iframe);
    
        // Create the download link
        var link = document.createElement('a');
        link.href = url;
        link.download = "molstar_my_filename.html";
        link.innerText = "Download HTML";
        link.style.display = "block";
        link.style.marginTop = "10px";
        wrapper.appendChild(link);
    }}, 100);
    """

    # Display the iframe
    display(HTML(f'<div id="{wrapper_id}"></div>'))
    display(Javascript(js_code))



import subprocess
from ovo.core.utils.resources import RESOURCES_DIR
from ovo.core.utils.tests import prepare_test_output_dir, TEST_SCHEDULER_KEY
import pandas as pd
import os


def test_scaffold_cli():
    output_dir = prepare_test_output_dir(__file__)
    cmd = [
        "ovo",
        "scheduler",
        "run",
        "rfdiffusion-end-to-end",
        output_dir,
        "--scheduler",
        TEST_SCHEDULER_KEY,
        "--max_memory",
        "4G",
        "--batch_size",
        "2",
        "--design_type",
        "scaffold",
        "--rfdiffusion_input_pdb",
        str((RESOURCES_DIR / "examples/inputs/5ELI_A.pdb").resolve()),
        "--rfdiffusion_num_designs",
        "1",
        "--rfdiffusion_contig",
        "A111-114/10/A117-119",
        "--rfdiffusion_run_parameters",
        "diffuser.T=1 inference.deterministic=True",
        "--mpnn_run_parameters",
        "--seed 42",
        "--refolding_tests",
        "af2_model_1_ptm_ft_1rec",
    ]
    subprocess.check_call(cmd)
    print("Saved to:", output_dir)
    subprocess.run(["ls", "-lR", output_dir.rstrip("/") + "/"], check=True)
    initial_guess = pd.read_json(os.path.join(output_dir, "contig1_batch1/af2_model_1_ptm_ft_1rec.jsonl"), lines=True)
    assert len(initial_guess) == 1



import glob
import os
import argparse
import json


REMARK_KEYS = [
    "Input contig",
    "Standardized contig",
    "Inpaint seq",
    "Chains",
    "Input hotspots",
    "Standardized hotspots",
]

REQUIRED_REMARK_KEYS = [
    "Standardized contig",
    "Chains",
]


def is_designed(contig_chain: str) -> bool:
    """
    Check if the contig chain is fixed or not.
    :param contig_chain: str, the contig chain string
    :return: bool, True if fixed, False otherwise
    """
    for segment in contig_chain.removesuffix("/0").split("/"):
        if not segment[0].isalpha():
            return True
    return False


def get_remark_header(pdb_path: str) -> tuple[str, list[str]]:
    """
    Get the REMARK header from the PDB file.
    :param pdb_path: str, path to the PDB file
    :return: str, REMARK header
    """
    with open(pdb_path, "r") as f:
        lines = f.readlines()

    all_remark_lines = []
    remark_header = ""
    for line in lines:
        if line.startswith("REMARK   1"):
            all_remark_lines.append(line.strip().removeprefix("REMARK   1 "))
            remark_header += line

    return remark_header, all_remark_lines


def parse_remark_lines(stripped_remark_lines: list[str]) -> dict[str, str]:
    """
    :param stripped_remark_lines:
        str, lines from the PDB file header, stripped of the "REMARK   1" prefix
    :return: dict, parsed remarks

    Example: example header to read
        REMARK   1 Input contig: "A45-46/10-15/A45-46/0 B24-26/5/B24-26/0 "
        REMARK   1 Input contig: "C10-20"
        REMARK   1 Standardized contig: "A45-46/13-13/A45-46/0 B24-26/5-5/B24-26/"
        REMARK   1 Standardized contig: "0 C10-20/0"
        REMARK   1 Chains: "A B C"
        REMARK   1 Input hotspots:
        REMARK   1 Standardized hotspots:
    """

    parsed_remark = {}
    for key in REMARK_KEYS:
        # Get lines that begin with this key
        remark_lines = [line.removeprefix(f"{key}:").strip() for line in stripped_remark_lines if line.startswith(key)]
        if not remark_lines and key in REQUIRED_REMARK_KEYS:
            raise ValueError(f"Missing REMARK line for key: {key}")

        values = []
        for line in remark_lines:
            if not line:
                if key in REQUIRED_REMARK_KEYS:
                    raise ValueError(f"Empty JSON value after key {key} in PDB file REMARK header")
                continue

            values.append(json.loads(line))

        parsed_remark[key] = "".join(values)

    return parsed_remark


def parse_inpaint_seq(inpaint_seq: str) -> list[tuple[str, int]]:
    """Parse A2-4/A10-12 to [("A", 2), ("A", 3), ("A", 4), ("A", 10), ("A", 11), ("A", 12)]"""
    if not inpaint_seq.strip():
        return []
    inpaint_s_list = []
    for i in inpaint_seq.strip().split("/"):
        if "-" in i:
            inpaint_s_list.extend([(i[0], p) for p in range(int(i.split("-")[0][1:]), int(i.split("-")[1]) + 1)])
        else:
            inpaint_s_list.append((i[0], int(i[1:])))
    return inpaint_s_list


def get_designed_residues(remark_dict: dict[str, str]) -> str:  # I can use remark_dict as input of this function here
    """
    Get the designed residues from the PDB file.
    :Arg:
        remark_dict: dict, parsed remarks
    :Return:
        designed_residues: str, designed residues in the format "A1 A2 A3 ..."
    Steps:
        - Read remarks lines from the PDB file header;
        - get designed residues to be written in the input json for LigandMPNN;
    """

    contigs_str = remark_dict.get("Standardized contig", None)
    chains_str = remark_dict.get("Chains", None)
    inpaint_seq = remark_dict.get("Inpaint seq", None)
    assert contigs_str, "No Standardized contigs found."
    assert chains_str, "No Standardized contigs found."

    contig_chains = contigs_str.split(" ")
    chains = chains_str.split(" ")

    # For example [("A", 2), ("A", 3), ("A", 4), ("A", 10), ("A", 11), ("A", 12)]
    parsed_inpaint_seq = set(parse_inpaint_seq(inpaint_seq))

    assert len(contig_chains) == len(chains), (
        f"Number of contigs and chains do not match: {len(contig_chains)} != {len(chains)}"
    )

    designed_residues = []
    for idx_chain, contig_chain in enumerate(contig_chains):
        chain = chains[idx_chain]

        i = 1
        assert " " not in contig_chain

        for segment in contig_chain.removesuffix("/0").split("/"):
            if is_designed(segment):
                start_len, end_len = map(int, segment.split("-"))
                assert start_len == end_len, (
                    f"Start len and end len of design segment should coincide: "
                    f"{start_len} != {end_len} in {contig_chain} "
                )

                for _ in range(start_len):
                    designed_residues.append(f"{chain}{i}")
                    i += 1

            else:
                input_chain = segment[0]
                # A10-12 -> 10, 12
                start, end = map(int, segment[1:].split("-"))
                for input_resnum in range(start, end + 1):
                    if (input_chain, input_resnum) in parsed_inpaint_seq:
                        # Fixed structure but inpainted (masked) sequence -> design this residue
                        designed_residues.append(f"{chain}{i}")
                    i += 1

    return " ".join(designed_residues)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--pdb_dir", type=str, help="Input directory with PDB files")
    parser.add_argument("--pdb_ids_json", type=str, help="Output json filename for multiple PDB input")
    parser.add_argument(
        "--redesigned_residues_json", type=str, help="Output json filename to redesign specific residues"
    )
    parser.add_argument(
        "--remark_json", default="./remark_multi.json", help="Output json filename to save REMARK lines from PDB files"
    )

    options = parser.parse_args()

    if os.path.isdir(options.pdb_dir):
        paths = sorted(glob.glob(os.path.join(options.pdb_dir, "*.pdb")))
        # paths = [pdb_path.replace("_standardized", "") for pdb_path in paths]
        print(f"Reading sequences from {len(paths):,} PDBs")
        pdb_path_ids = {pdb_path: "" for pdb_path in paths}

        remark_header_by_pdb_path = {}
        remark_lines_by_pdb_path = {}
        for pdb_path in paths:
            remark_header, remark_lines = get_remark_header(pdb_path)
            if not remark_header:
                raise ValueError(
                    f'This script requires a "standardized" PDB output with the REMARK header '
                    f"(as produced by the ovo rfdiffusion-backbone workflow), "
                    f"no header found in: {pdb_path}"
                )

            remark_header_by_pdb_path[pdb_path] = remark_header
            remark_lines_by_pdb_path[pdb_path] = remark_lines

        redesigned_residues_by_pdb_path = {
            pdb_path: get_designed_residues(parse_remark_lines(remark_lines))
            for pdb_path, remark_lines in remark_lines_by_pdb_path.items()
        }
    else:
        raise ValueError("Input must be a directory with PDB files")

    # Write pdb ids json
    with open(options.pdb_ids_json, "w") as f:
        json.dump(pdb_path_ids, f)
    print("Saved pdb ids to:", options.pdb_ids_json)

    # Write REMARK lines per pdb
    with open(options.remark_json, "w") as f:
        json.dump(remark_header_by_pdb_path, f)
    print("Saved remark lines to:", options.remark_json)

    # Write redesigned residues json
    with open(options.redesigned_residues_json, "w") as f:
        json.dump(redesigned_residues_by_pdb_path, f)
    print("Saved redesigned residues to:", options.redesigned_residues_json)



from Bio import PDB


class PDBSelector(PDB.Select):
    """Class helper that selects a specific chain in a PDB file."""

    def __init__(self, chain_id, start_residue=None, end_residue=None):
        self.chain_id = chain_id
        self.start_residue = start_residue
        self.end_residue = end_residue

    def accept_model(self, model):
        return True  # Accept all models

    def accept_chain(self, chain):
        return chain.id == self.chain_id

    # TODO: Handle insertion codes
    def accept_residue(self, residue):
        # Accept residue if its id[1] falls within the specified range
        if self.start_residue and self.end_residue:
            return self.start_residue <= residue.id[1] <= self.end_residue
        else:
            return True

    def accept_atom(self, model):
        return True  # Accept all atoms


def get_resname_resnum_atomname_atomnum_str(atom_ppdb, residue_indices: list | None) -> str | None:
    """Return a string of resname_resnum_atomname_atomnum for each atom in the list of residue_indices."""
    if not residue_indices:
        return None

    interfaces_residues_str = ",".join(
        atom_ppdb.loc[residue_indices].apply(
            lambda row: "_".join([row.residue_name, str(row.residue_number), row.atom_name, str(row.atom_number)]),
            axis=1,
        )
    )
    return interfaces_residues_str



import os
from copy import deepcopy

import streamlit as st

from ovo.app.components.molstar_custom_component import molstar_custom_component, StructureVisualization
from ovo.app.components.molstar_custom_component.dataclasses import ContigSegment
from ovo.core.database import WorkflowTypes, Workflow
from ovo.core.utils.residue_selection import parse_selections, from_segments_to_hotspots
from ovo.core.utils.formatting import safe_filename
from ovo.core.utils.pdb import get_pdb, filter_pdb_str

from ovo import storage
from ovo.core.utils.pdb import add_glycan_to_pdb


def initialize_workflow(page_key: str, workflow_name: str, include_subclasses: bool = False) -> Workflow:
    """Initialize a workflow in the session state based on dropdown with available workflow variants (subclasses).

    :param page_key: Key identifying the workflow page, used as key in session_state.workflows
    :param workflow_name: Base name of the workflow (e.g. "rfdiffusion-end-to-end")
    :param include_subclasses: Whether to let user choose from subclasses of the base workflow as variants
    """
    variant_names = WorkflowTypes.get_subclass_names(workflow_name) if include_subclasses else [workflow_name]
    if len(variant_names) > 1:
        with st.container(horizontal=True):
            workflow = st.session_state.workflows.get(page_key, None)
            variant_name = st.selectbox(
                "Workflow variant",
                options=variant_names,
                index=variant_names.index(workflow.name) if workflow else 0,
                width=450,
                key=f"workflow_variant_{page_key}",
            )
    else:
        variant_name = variant_names[0]

    if not st.session_state.workflows.get(page_key) or st.session_state.workflows[page_key].name != variant_name:
        # If workflow is not initialized yet or variant has changed, (re)initialize it
        # TODO carry over parameters when changing variant if possible
        st.session_state.workflows[page_key] = WorkflowTypes.get(variant_name)()
        # Set a copy of the initial workflow to be able to reset parameters later
        st.session_state.initial_workflows[page_key] = deepcopy(st.session_state.workflows[page_key])

    return st.session_state.workflows[page_key]


def pdb_input_component(old_pdb_code: str | None) -> tuple[str, bytes] | None:
    st.write("Enter PDB code, UniProt ID or upload your own structure file")
    with st.container(horizontal=True, vertical_alignment="bottom"):
        new_pdb_code = st.text_input(
            "Download by ID",
            placeholder="Input PDB code or UniProt ID...",
            value=old_pdb_code,
            key="input_pdb_code",
            width=270,
        )
        st.button("Confirm")

    uploaded_file = st.file_uploader("...or upload a PDB file", type=["pdb", "pdb1"], key="input_pdb_file")

    # Check if a file has just been uploaded
    if uploaded_file is not None:
        pdb_input_bytes = uploaded_file.getvalue()
        filename, _ = os.path.splitext(safe_filename(uploaded_file.name))
        return filename, pdb_input_bytes

    elif new_pdb_code and new_pdb_code != old_pdb_code:
        with st.spinner(f"Downloading PDB {new_pdb_code}"):
            return new_pdb_code, get_pdb(new_pdb_code)

    return None


@st.fragment
def sequence_selection_fragment(
    page_key: str,
    input_name: str,
    color="chain-id",
    representation_type="cartoon+ball-and-stick",
    write_segments: bool = True,
    fixed_segments: list[ContigSegment] | None = None,
    **selection_kwargs,
):
    """Fragment to display input structure and allow residue selection via molstar component.

    :param page_key: Key identifying the workflow page, typically __file__
    :param input_name: Name of the input structure, used for download filename
    :param color: Initial color scheme for molstar component
    :param representation_type: Initial representation type for molstar component
    :param write_segments: Whether to write selected segments (True) or individual residues (False)
    :param fixed_segments: If provided, only these segments will be shown in the structure (others hidden)
    :param selection_kwargs: Additional kwargs passed to workflow.get/set_selected_segments methods
    """

    workflow = st.session_state.workflows[page_key]

    if page_key not in st.session_state.selection_history:
        st.session_state.selection_history[page_key] = {}

    if input_name not in st.session_state.selection_history[page_key]:
        st.session_state.selection_history[page_key][input_name] = [[]]

    selection_history = st.session_state.selection_history[page_key][input_name]

    pdb_input_string = storage.read_file_str(workflow.get_input_pdb_path())

    if fixed_segments and not st.toggle(
        "Show full input structure", help="By default, only the fixed input segments are shown."
    ):
        pdb_input_string = filter_pdb_str(
            pdb_input_string,
            segments=[seg.value for seg in fixed_segments],
            add_ter=False,  # do not add TER so that molstar shows the sequence all in one piece
        )

    left, mid, right = st.columns([1.5, 1.5, 3])
    with left:
        if st.toggle(
            "Show glycosylation sites",
            key="show_glycosylation_sites_input",
            help="Visualize glycosylation sites on the structure based on glycosylation motifs in the sequence (NXS, NXT, X = any amino acid except P). "
            "The positions of glycans are approximate, based on the 5eli NAG template, "
            "nonexistent bonds might be displayed.",
        ):
            pdb_input_string, glycosylated_residues = add_glycan_to_pdb(pdb_input_string)
            if glycosylated_residues:
                st.warning(f"Glycosylated motif found at {','.join(glycosylated_residues)}")
            else:
                st.info("No glycosylation motif found.")

    with mid:
        representation_types = {
            "cartoon": "Cartoon",
            "cartoon+ball-and-stick": "Cartoon + Side chains",
            "molecular-surface": "Molecular Surface",
        }
        representation_type = st.selectbox(
            "Representation type",
            options=list(representation_types.keys()),
            index=list(representation_types.keys()).index(representation_type),
            format_func=lambda x: representation_types[x],
            key="representation_type_input",
            label_visibility="collapsed",
        )

    with right:
        colors = {
            "chain-id": "Color by chain",
            "hydrophobicity": "Color by hydrophobicity (green = hydrophobic, red = hydrophilic)",
            "residue-charge": "Color by residue charge (blue = positive, red = negative)",
        }
        color = st.selectbox(
            "Color scheme",
            options=list(colors.keys()),
            index=list(colors.keys()).index(color),
            format_func=lambda x: colors[x],
            key="color_scheme_input",
            label_visibility="collapsed",
        )

    with st.container(horizontal=True, vertical_alignment="center"):
        force_reload = False
        if (
            st.button(":material/undo: Undo selection", disabled=len(selection_history) <= 1)
            and len(selection_history) >= 2
        ):
            # Go one selection backwards
            selection_history.pop(-1)
            previous_value = selection_history.pop(-1)
            workflow.set_selected_segments(previous_value, **selection_kwargs)
            force_reload = True

        selection_container = st.empty()
        selection_container.write(
            "Click residues in sequence or structure to add them to the selection. Shift+Click to select a residue range."
        )

    selected_str = molstar_custom_component(
        structures=[
            StructureVisualization(
                pdb=pdb_input_string,
                highlighted_selections=workflow.get_selected_segments(**selection_kwargs),
                color=color,
                contigs=fixed_segments if color == "chain-id" else None,
                representation_type=representation_type,
            )
        ],
        selection_mode=True,
        show_controls=True,
        height=700,
        download_filename=input_name,
        key="input",
        force_reload=force_reload,
    )

    if selected_str is None:
        # component has not been initialized yet, stop here
        return

    selection = parse_selections(selected_str)

    workflow.set_selected_segments(selection, **selection_kwargs)

    # Added Logic to update selection history. Only add to history if the selection
    # - is not empty ([], it's output every other re-rendering)
    # - is not the current selection already (to avoid duplicates due to multiple re-rendering)
    # FIXME: Only case not handled is when, at start, one single residue only is selected and try to undo the selection
    if not force_reload and selection and (not selection_history or selection != selection_history[-1]):
        selection_history.append(selection)

    if selection:
        if write_segments:
            selection_container.write(f"Selected segments: {'/'.join(selection)}")
        else:
            residues = from_segments_to_hotspots(selection)
            selection_container.write(f"Selected residues: {residues}")



import itertools
import json
from collections import defaultdict
from io import StringIO
from typing import List, Dict
from Bio.PDB import PDBParser


def from_residues_to_segments(
    chain_id: str, residues: list[int], start_res: int | None = None, end_res: int | None = None
) -> list[str]:
    """Convert a list of residues [3,4,5,6,9,10,...] into a list of segments [3-6,9-10,...].

    Args:
        chain_id (str): Required. The chain ID.
        residues (list[int]): Required. A list of residues.
        start_res (int, optional): The start residue. If specified, trims residues from start_res. Defaults to None.
        end_res (int, optional): The end residue. If specified, trims residues to end_res. Defaults to None.

    Returns:
        list(str): A list of segments, i.e. [3-6,9-10,...]
    """

    if not residues:
        return []

    if start_res is not None:
        residues = [res for res in residues if res >= start_res]
    if end_res is not None:
        residues = [res for res in residues if res <= end_res]

    segments = []
    start = end = residues[0]
    for i in range(1, len(residues)):
        if residues[i] == residues[i - 1] + 1:
            end = residues[i]
        else:
            segments.append(f"{chain_id}{start}-{end}")
            start = end = residues[i]
    segments.append(f"{chain_id}{start}-{end}")
    return segments


def from_residues_to_hotspots(chain_id: str, residues: list[int]):
    """Convert a chain_id (i.e. A) and list of residues [3,4,...] into list of hotspots [A3, A4, ...]."""
    return [f"{chain_id}{res}" for res in residues]


def parse_selections(selections: str) -> list[str]:
    """
    Input: a JSON-encoded representation of the selections that looks like this
        {"sequenceSelections":[{"chainId":"A","residues":[10,11,12,13,15,16,17,18,19,20"]}]}
    Outputs list o segments ["A10-13", "A15-20"]
    """

    if selections is None:
        return []

    selections_json = json.loads(selections)
    selections_list = selections_json["sequenceSelections"]

    # for each of these, let's create a different representation
    result = []
    for selection in selections_list:
        chain_id = selection["chainId"]
        residues = selection["residues"]
        result += from_residues_to_segments(chain_id, residues)

    return result


def get_chains_and_contigs(pdb_str: str | None) -> Dict[str, str] | None:
    if pdb_str is None:
        return None

    parser = PDBParser(QUIET=True)

    struc = parser.get_structure("PDB_structure", StringIO(pdb_str))

    segments_by_chain = {}
    # Iterate through models, chains, and residues
    for model in struc:
        for chain in model:
            chain_id = chain.id  # Get the chain identifier
            residues = set()  # Use a set to hold unique residues

            for residue in chain:
                # Ensure the residue is an ATOM or ALA/GLY/other relevant residue
                if residue.id[0] == " " and int(residue.id[1]) > -1:
                    residues.add(int(residue.get_id()[1]))  # Get the residue number (not the insertion code)

            segments_by_chain[chain_id] = sorted(residues)

    segments_by_chain_json = {
        chain: parse_selections(json.dumps({"sequenceSelections": [{"chainId": chain, "residues": list(residues)}]}))
        for chain, residues in segments_by_chain.items()
    }

    return {chain: "/".join(segments) for chain, segments in segments_by_chain_json.items()}


def from_hotspots_to_segments(hotspots: str) -> list[str] | None:
    hotspots_by_chain = defaultdict(list)

    if not hotspots:
        return None

    # TODO: Maybe some more regex checks?
    for s in hotspots.split(","):
        hotspots_by_chain[s[0]].append(int(s[1:].replace(" ", "")))

    result = []
    for chain, hotspot_nums in hotspots_by_chain.items():
        hotspot_nums.sort()
        result.extend(from_residues_to_segments(chain, hotspot_nums))

    return result


def from_segments_to_hotspots(segments: list[str] | None) -> str:
    """Convert a list of segments ["A3-5","A9-10",...] to a STRING of hotspots "A3,A4,A5,A9,A10"."""

    if not segments:
        return ""

    hotspots = []
    for segment in segments:
        chain = segment[0]
        start_resnum = int(segment[1:].split("-")[0])
        end_resnum = int(segment[1:].split("-")[-1])
        hotspots.extend([f"{chain}{resnum}" for resnum in range(start_resnum, end_resnum + 1)])

    return ",".join(hotspots)


def from_contig_to_residues(chain_segment: str | None) -> list[int] | None:
    """Convert a str chain_segment / contig "A3-5/A9-10" to a list of residues [3,4,5,9,10]."""

    if not chain_segment:
        return None

    contig_segments = chain_segment.split(" ")[0].removesuffix("/0").split("/")

    residues = []
    # Merge non-consecutive residues into a list
    for segment in contig_segments:
        start_resnum = int(segment[1:].split("-")[0])
        end_resnum = int(segment[1:].split("-")[-1])
        residues += list(range(start_resnum, end_resnum + 1))

    return residues


def from_residues_to_chain_breaks(residues: list[int]) -> List[str] | None:
    residues.sort()
    chain_break_segments = []
    for i in range(1, len(residues)):
        # Check for a gap between consecutive numbers
        if residues[i] - residues[i - 1] > 1:
            missing_start_res = residues[i - 1]
            missing_end_res = residues[i]
            chain_break_segments.append(f"{missing_start_res}-{missing_end_res}")

    return chain_break_segments


def parse_partial_diffusion_binder_contig(binder_contig: str) -> tuple[int, list[str]]:
    """Get binder length and DESIGNED segments from binder contig as they correspond to positions in binder chain A

    default case for a peptide of 12 residues is simply binder_contig="12-12" turned into ["A1-12"]
    or when redesigning only A1 and A3-7 in a peptide of 12 residues, this would be
    binder_contig="1-1/A2-2/5-5/A8-12" turned into ["A1-1", "A3-7"]

    :param binder_contig: binder contig, e.g. 12-12 or 1-1/A2-2/5-5/A8-12
    :return: binder length (int) and list of designed segments (list of str, e.g. ["A1-1", "A3-7"])
    """
    assert " " not in binder_contig, f"Expected single chain in binder_contig, got: {binder_contig}"
    pos = 0
    designed_segments = []
    for segment in binder_contig.split("/"):
        if segment[0].isalpha():
            # fixed segment
            assert segment[0] == "A", f"Expected first contig in partial diffusion to be in chain A, got: {segment}"
            assert "-" in segment, f"Expected fixed segment in partial diffusion to be in format A2-3, got: {segment}"
            start, end = segment[1:].split("-")
            assert int(start) <= int(end), (
                f"Invalid fixed segment in partial diffusion, start must be <= end, got: {segment}"
            )
            assert int(start) == pos + 1, (
                f"Segment {segment} at generated position {pos + 1} should start with A{pos + 1} in partial diffusion"
            )
            pos += int(end) - int(start) + 1
        else:
            # designed segment
            assert "-" in segment, f"Expected designed segment in partial diffusion to be in format 5-5, got: {segment}"
            segment_length, _ = segment.split("-")
            assert segment_length == _, (
                f"Designed segments should have same start and end residue in partial diffusion, got: {segment}"
            )
            designed_segments.append(f"A{pos + 1}-{pos + int(segment_length)}")
            pos += int(segment_length)
    return pos, designed_segments


def create_partial_diffusion_binder_contig(redesigned_segments: list[str], binder_length: int) -> str:
    """Create a new binder contig for partial diffusion from the designed segments and old binder contig

    default case for a peptide of 12 residues:
    redesigned_segments=[] -> binder_contig="12-12"
    or when redesigning only A1 and A3-7 in a peptide of 12 residues,
    redesigned_segments=["A1-1", "A3-7"] -> binder_contig="1-1/A2-2/5-5/A8-12"

    :param redesigned_segments: list of designed segments (list of str, e.g. ["A1-1", "A3-7"])
    :param binder_length: length of binder (int)
    :return: new binder contig (str)
    """
    assert all([segment[0] == "A" for segment in redesigned_segments]), (
        f"Expected all redesigned segments to be in chain A, got: {redesigned_segments}"
    )
    redesigned_str = from_segments_to_hotspots(redesigned_segments)
    redesigned_positions = [int(res[1:]) for res in redesigned_str.split(",")] if redesigned_str else []
    all_positions = list(range(1, binder_length + 1))
    new_binder_segments = []
    for is_redesigned, positions in itertools.groupby(all_positions, lambda x: x in redesigned_positions):
        positions = list(positions)
        if is_redesigned:
            new_binder_segments.append(f"{len(positions)}-{len(positions)}")
        else:
            new_binder_segments.append(f"A{positions[0]}-{positions[-1]}")

    return "/".join(new_binder_segments)



import gzip
import json

import numpy as np
import os
import requests

from biopandas.pdb import PandasPdb

import pandas as pd
from copy import deepcopy
from Bio import PDB, Align
from Bio.PDB import PDBParser
from Bio.SeqUtils import seq1

from io import StringIO
from scipy.spatial.transform import Rotation as R


aa3to1 = {
    "CYS": "C",
    "ASP": "D",
    "SER": "S",
    "GLN": "Q",
    "LYS": "K",
    "ILE": "I",
    "PRO": "P",
    "THR": "T",
    "PHE": "F",
    "ASN": "N",
    "GLY": "G",
    "HIS": "H",
    "LEU": "L",
    "ARG": "R",
    "TRP": "W",
    "ALA": "A",
    "VAL": "V",
    "GLU": "E",
    "TYR": "Y",
    "MET": "M",
}


# Pre-calculated transformed displacements using 5eli NAG as template in the CB-ND2 z-axis reference frame
transformed_displacements = np.array(
    [
        [-0.4082622, -0.45912408, -1.30803404],
        [-0.70450961, 0.68773313, -2.28396899],
        [-0.22737329, 0.32573119, -3.6814555],
        [1.27596815, 0.09700733, -3.66800543],
        [1.63672552, -0.96640494, -2.63479564],
        [2.61316972, -0.48611626, -1.5859937],
        [-2.5847387, 2.26184417, -2.25385266],
        [-4.0755428, 2.41315718, -2.28422532],
        [-2.12259706, 1.0096836, -2.29977646],
        [-0.55055668, 1.37749703, -4.58249662],
        [1.71515412, -0.32738686, -4.95309542],
        [0.46600102, -1.44944231, -1.94881529],
        [3.87324501, -1.12614701, -1.72277423],
        [-1.83356036, 3.2290743, -2.18745253],
    ]
)


def fix_contigs(contigs, parsed_pdb):
    def fix_contig(contig):
        INF = float("inf")
        X = contig.split("/")
        Y = []
        for n, x in enumerate(X):
            if x[0].isalpha():
                C, x = x[0], x[1:]
                S, E = -INF, INF
                if x.startswith("-"):
                    E = int(x[1:])
                elif x.endswith("-"):
                    S = int(x[:-1])
                elif "-" in x:
                    (S, E) = (int(y) for y in x.split("-"))
                elif x.isnumeric():
                    S = E = int(x)
                new_x = ""
                c_, i_ = None, 0
                for c, i in parsed_pdb["pdb_idx"]:
                    if c == C and i >= S and i <= E:
                        if c_ is None:
                            new_x = f"{c}{i}"
                        else:
                            if c != c_ or i != i_ + 1:
                                new_x += f"-{i_}/{c}{i}"
                        c_, i_ = c, i
                Y.append(new_x + f"-{i_}")
            elif "-" in x:
                # sample length
                s, e = x.split("-")
                m = np.random.randint(int(s), int(e) + 1)
                Y.append(f"{m}-{m}")
            elif x.isnumeric() and x != "0":
                Y.append(f"{x}-{x}")
        return "/".join(Y)

    return [fix_contig(x) for x in contigs]


def get_pdb(pdb_code: str) -> bytes:
    if len(pdb_code) == 4:
        base_url = f"https://files.rcsb.org/download/{pdb_code}.pdb"
    else:
        base_url = f"https://alphafold.ebi.ac.uk/files/AF-{pdb_code.upper()}-F1-model_v6.pdb"

    # NO_VERIFY_SSL is set to true in gh action for testing purposes (problem with self-signed certificate)
    response = requests.get(base_url, verify=not os.getenv("NO_VERIFY_SSL", False))

    if not response.ok:
        raise ValueError(f"Error downloading PDB {pdb_code}")
    if base_url.endswith(".gz"):
        return gzip.decompress(response.content)
    return response.content


class PDBSegmentSelector(PDB.Select):
    def __init__(self, segments: list[str]):
        """Select only the specified segments from a PDB structure.

        :param: segments: list of str with the segments to select, e.g. ["A5", "A10-20", "B30-40", "C"]
        """
        self.parsed_segments = []
        for segment in segments:
            chain_id = segment[0]
            assert chain_id.isalpha(), f"Invalid chain ID in segment, expected A123-456, got: {segment}"
            if len(segment) == 1:
                # full chain
                self.parsed_segments.append((chain_id, 1, np.inf))
            elif "-" in segment:
                # residue range
                start_residue, end_residue = segment[1:].split("-")
                self.parsed_segments.append((chain_id, int(start_residue), int(end_residue)))
            else:
                # single residue
                residue_number = int(segment[1:])
                self.parsed_segments.append((chain_id, residue_number, residue_number))

    def accept_model(self, model):
        return True  # Accept all models

    def accept_chain(self, chain):
        return chain.id in [seg[0] for seg in self.parsed_segments]  # Pre-filter chains

    def accept_residue(self, residue):
        residue_chain_id = residue.full_id[2]
        residue_number = residue.id[1]
        return any(
            residue_chain_id == chain_id and (start <= residue_number <= end)
            for chain_id, start, end in self.parsed_segments
        )

    def accept_atom(self, model):
        return True  # Accept all atoms


def detect_glycosylation_sites(
    atom_ppdb: pd.DataFrame,
    chains: list[str] | str | None = None,
    query_atoms: list[str] | None = None,
    cyclic: bool = False,
) -> dict | None:
    """Find glycosylation sites in the PDB file
    :param:
        atom_ppdb: Pandas DataFrame with the ATOM records of the PDB file
        chains: str or list of str with the chain IDs to search for glycosylation sites
        query_atoms : list of str with the atom names to search for glycosylation sites

    :return:
        glycosylation_dict: dictionary with the coordinates of the query glycosylated atoms
    """
    glycosylation_pattern = ["ASN", "* except PRO", ["THR", "SER"]]  # Define the pattern N*S N*T

    if query_atoms is None:
        query_atoms = ["CA", "C", "CB", "ND2"]

    if isinstance(chains, str):
        chains = [chains]
    elif chains is None:
        chains = atom_ppdb["chain_id"].unique()

    atom_ppdb = atom_ppdb.query("chain_id in @chains")

    residues = atom_ppdb[["chain_id", "residue_number", "residue_name"]].drop_duplicates().set_index("residue_number")

    # Create a DataFrame of overlapping residue triplets (e.g. resnum
    # Before dropna three_mers looks like this:
    # num chain res[num] res[num+1] res[num+2]
    # 129   A     ASN       THR        THR
    # 130   A     THR       THR        VAL
    # 131   A     THR       VAL        PHE
    # 132   A     VAL       PHE        None
    # 132   A     PHE       None       None

    three_mers = pd.concat(
        [residues.chain_id] + [residues.residue_name.shift(-i) for i in range(len(glycosylation_pattern))], axis=1
    )

    three_mers.columns = ["chain_id", "resname1", "resname2", "resname3"]

    if cyclic:
        three_mers.loc[three_mers.index[-1], "resname2"] = three_mers.iloc[0].loc["resname1"]
        three_mers.loc[three_mers.index[-1], "resname3"] = three_mers.iloc[0].loc["resname2"]
        three_mers.loc[three_mers.index[-2], "resname3"] = three_mers.iloc[0].loc["resname1"]

    three_mers = three_mers.dropna()

    # Check for the glycosylation pattern
    is_glycosylated = three_mers.apply(
        lambda row: (row.iloc[1] == glycosylation_pattern[0])
        and (row.iloc[2] != "PRO")
        and (row.iloc[3] in glycosylation_pattern[2]),
        axis=1,
    )

    glycosylated_residues = three_mers[is_glycosylated]

    if glycosylated_residues.empty:
        return None

    glycosylated_chains = glycosylated_residues.chain_id.unique()
    glycosylation_dict = {}
    for chain in glycosylated_chains:
        chain_atom_ppdb = atom_ppdb.query("chain_id==@chain")
        for resnum in glycosylated_residues[glycosylated_residues.chain_id == chain].index:
            glycosylation_dict[(chain, resnum)] = {}
            for atom in query_atoms:
                res_coords_df = chain_atom_ppdb.query("residue_number==@resnum")
                glycosylation_dict[(chain, resnum)][atom] = res_coords_df.query("atom_name==@atom")[
                    ["x_coord", "y_coord", "z_coord"]
                ].values.flatten()

    return glycosylation_dict


def calculate_coords_from_transformed_displacements(P1, P2) -> np.ndarray:
    """Calculate the coordinates of the glycan atoms based on the transformed displacements

    It uses the pre-calculated transformed displacements to calculate the coordinates of the glycan atoms

    :param: P1, P2: 3D coordinates in the PDB reference frame of the ND2 and CB atoms of the glycosylated residue
    :return: 3D coordinates in the PDB reference frame of the glycan atoms
    """

    # Convert points to numpy arrays
    P1 = np.array(P1)
    P2 = np.array(P2)

    # Vector from P1 to P2
    V = P2 - P1
    U = V / np.linalg.norm(V)  # Normalize

    # Z-axis unit vector
    z_axis = np.array([0, 0, 1])

    # Calculate the rotation matrix aligning U to the Z-axis
    rotation_matrix = R.align_vectors([z_axis], [U])[0].as_matrix()

    # Inverse rotation
    inverse_rotation = rotation_matrix.T

    # Compute inverse translation
    inverse_translation = -inverse_rotation @ P1  # Inverse translation based on the rotation

    # Create the inverse transformation matrix
    inverse_matrix = np.eye(4)
    inverse_matrix[:3, :3] = inverse_rotation
    inverse_matrix[:3, 3] = inverse_translation

    original_displacements = []
    for displacement in transformed_displacements:
        # Convert to homogeneous coordinates
        displacement_homogeneous = np.array([*displacement, 0])  # The last element is 0 for a displacement
        # Apply inverse transformation matrix
        original_displacement_homogeneous = inverse_matrix @ displacement_homogeneous
        original_displacement = original_displacement_homogeneous[:3]  # Get x, y, z
        original_displacements.append(tuple(original_displacement))

    coords = np.array(original_displacements) + P1

    return coords


def add_glycan_to_pdb(pdb_str: str) -> tuple[str, list[str] | None]:
    """Add glycan atoms to the PDB string"""

    # Read PDB file as df from string
    ppdb = PandasPdb().read_pdb_from_list(pdb_str.splitlines(True))

    # calculate glycosylated sites per (chain, residue)
    glycosylated_residues = detect_glycosylation_sites(ppdb.df["ATOM"])

    if not glycosylated_residues:
        return pdb_str, None

    glycosylated_residues_str = [f"{chain_resnum[0]}{chain_resnum[1]}" for chain_resnum in glycosylated_residues]

    # calculate coords for glycanes sites per (chain, residue)
    coords_to_insert_glycane_per_resnum = {}
    for chain_and_resnum, atoms_coords_dict in glycosylated_residues.items():
        ND2_coords, CB_coords, CA_coords, C_coords = (
            atoms_coords_dict["ND2"],
            atoms_coords_dict["CB"],
            atoms_coords_dict["CA"],
            atoms_coords_dict["C"],
        )

        # This allows to visualize the glycan atoms also if ND2, CB are missing
        P1, P2 = None, None
        if C_coords.size > 0 and CA_coords.size > 0:
            P1, P2 = C_coords, CA_coords
        if CA_coords.size > 0 and CB_coords.size > 0:
            P1, P2 = CB_coords, CA_coords
        if ND2_coords.size > 0 and CB_coords.size > 0:
            P1, P2 = ND2_coords, CB_coords

        if P1 is not None and P2 is not None:
            coords_to_insert_glycane_per_resnum[chain_and_resnum] = calculate_coords_from_transformed_displacements(
                P1, P2
            )

    # Create a PandasPdb object to store the glycan atoms
    atom_names_nag = ["C1", "C2", "C3", "C4", "C5", "C6", "C7", "C8", "N2", "O3", "O4", "O5", "O6", "O7"]

    N_rows = len(atom_names_nag) * len(coords_to_insert_glycane_per_resnum)

    glycan_ppdb = deepcopy(ppdb)

    ppdb_hetatm = pd.DataFrame()

    ppdb_hetatm["record_name"] = ["HETATM"] * N_rows
    ppdb_hetatm["atom_number"] = list(range(1, N_rows + 1))
    ppdb_hetatm["blank_1"] = [""] * N_rows
    ppdb_hetatm["alt_loc"] = [""] * N_rows
    ppdb_hetatm["residue_name"] = ["NAG"] * N_rows
    ppdb_hetatm["blank_2"] = [""] * N_rows
    ppdb_hetatm["insertion"] = [""] * N_rows
    ppdb_hetatm["blank_3"] = [""] * N_rows
    ppdb_hetatm["atom_name"] = atom_names_nag * len(coords_to_insert_glycane_per_resnum)

    chain_id = []
    residue_number = []
    x_coords, y_coords, z_coords = [], [], []
    element_symbol = []
    for nag_resnum, ((NAG_chain, resnum), glycan_atoms) in enumerate(coords_to_insert_glycane_per_resnum.items()):
        chain_id += [NAG_chain] * len(atom_names_nag)
        residue_number += [int(nag_resnum + 1)] * len(atom_names_nag)
        x_coords += list(glycan_atoms[:, 0])
        y_coords += list(glycan_atoms[:, 1])
        z_coords += list(glycan_atoms[:, 2])
        element_symbol += ["C"] * 8 + ["N"] + ["O"] * 5

    ppdb_hetatm["chain_id"] = chain_id
    ppdb_hetatm["residue_number"] = residue_number
    ppdb_hetatm["x_coord"] = x_coords
    ppdb_hetatm["y_coord"] = y_coords
    ppdb_hetatm["z_coord"] = z_coords
    ppdb_hetatm["element_symbol"] = element_symbol

    ppdb_hetatm["occupancy"] = [1.0] * N_rows
    ppdb_hetatm["b_factor"] = [1.0] * N_rows
    ppdb_hetatm["blank_4"] = [""] * N_rows
    ppdb_hetatm["segment_id"] = [""] * N_rows
    ppdb_hetatm["charge"] = [np.nan] * N_rows
    ppdb_hetatm["line_idx"] = [np.nan] * N_rows

    if ppdb_hetatm.empty:
        return pdb_str, None

    glycan_ppdb.df["HETATM"] = ppdb_hetatm.loc[:, ppdb.df["ATOM"].columns]

    # Save the glycan PDB to a string
    pdb_glycan_string = glycan_ppdb.to_pdb_stream()

    return pdb_glycan_string.getvalue(), glycosylated_residues_str


def get_atom_coordinates(
    structure: PDB.Structure.Structure,
    chain_id: str | None,
    residues: list[int] | None,
    all_atom: bool = False,
    model_index=0,
) -> tuple[list[dict[str, np.ndarray]], list[PDB.Residue.Residue]]:
    coords = []
    res_final = []

    model = structure[model_index]
    for chain in model:
        if chain_id and chain.id != chain_id:
            continue
        for residue in chain:
            if residues and residue.id[1] not in residues:
                continue
            coords_dict = {}
            for atom in residue:
                if all_atom or atom.id == "CA":
                    coords_dict[atom.id] = atom.coord
            if coords_dict:
                coords.append(coords_dict)
                res_final.append(residue)
    return coords, res_final


def _align_sequences_and_get_indices(seqs: list[str]):
    """Aligns multiple protein sequences based on their amino acid sequences.
    Returns a tuple, a list of aligned sequences and list of indices for each sequence that correspond to the aligned residues."""
    aligner = Align.PairwiseAligner()
    aligner.open_gap_score = -7.0
    aligner.extend_gap_score = -2.0

    alignments = []
    aligned_seqs_indices = []

    # Process first sequence as the reference
    aligned_indexes_ref = []

    # Align each sequence to the reference (seqs[0])
    for i, seq in enumerate(seqs):
        if i == 0:
            # Skip aligning the reference to itself
            continue

        # Align current sequence to reference
        alignment = sorted(aligner.align(seqs[0], seq), key=lambda x: x.score, reverse=True)[0]
        alignments.append(alignment)

        # Extract aligned indices for current sequence
        aligned_indices = [idx for start, end in alignment.aligned[1] for idx in range(start, end)]
        aligned_seqs_indices.append(aligned_indices)

        # Extract aligned indices for reference sequence (only need to do once)
        if not aligned_indexes_ref:
            aligned_indexes_ref = [idx for start, end in alignment.aligned[0] for idx in range(start, end)]

    # Insert reference sequence alignment at the beginning
    aligned_seqs_indices.insert(0, aligned_indexes_ref)

    return aligned_seqs_indices


def align_multiple_proteins_pdb(
    pdb_strs: list[str],
    chain_residue_mappings: list[list[tuple[str, list[int] | None]] | None],
    force_sequence_alignment: bool = False,
    all_atom: bool = False,
    verbose: bool = False,
) -> tuple[list[str], float]:
    """Aligns multiple protein structures based on their atoms (CA or all).

    :param pdb_strs: list of PDB strings
    :param chain_residue_mappings: list of lists of tuples with chain ID and residues to align,
                                   if None provided, then whole chain/structure is aligned
    :param force_sequence_alignment: if True, always align based on sequence even if lengths match
    :param all_atom: if True, align using all atoms from matched residues (not just CA atoms)
    :param verbose: if True, print information about the alignment process
    """
    assert len(pdb_strs) == len(chain_residue_mappings), (
        f"Expected same number of structures and chain residue mappings, got {len(pdb_strs)} != {len(chain_residue_mappings)}"
    )

    parser = PDB.PDBParser(QUIET=True)
    structures = [parser.get_structure(f"Protein{i + 1}", StringIO(pdb_str)) for i, pdb_str in enumerate(pdb_strs)]
    coords_list = []
    residues_list = []

    for i, (structure, mappings) in enumerate(zip(structures, chain_residue_mappings)):
        structure_coords = []
        structure_residues = []

        if not mappings:
            coords, res_final = get_atom_coordinates(structure, None, None, all_atom=all_atom)
            if coords:
                coords_list.append(coords)
                residues_list.append(res_final)
                continue
            else:
                raise ValueError(f"No atoms found in structure {i + 1}")

        for chain_id, residues in mappings:
            coords, res_final = get_atom_coordinates(structure, chain_id, residues, all_atom=all_atom)
            if coords:
                structure_coords.extend(coords)
                structure_residues.extend(res_final)
            else:
                raise ValueError(f"No atoms found in structure {i + 1} for chain {chain_id} and residues {residues}")

        coords_list.append(structure_coords)
        residues_list.append(structure_residues)

    seqs = ["".join([aa3to1.get(residue.resname, "X") for residue in residues]) for residues in residues_list]

    sequence_alignment = force_sequence_alignment or any(len(seqs[i]) != len(seqs[0]) for i in range(1, len(seqs)))
    if sequence_alignment:
        aligned_seqs_indices = _align_sequences_and_get_indices(seqs)
    else:
        aligned_seqs_indices = [list(range(len(seqs[0])))] * len(seqs)

    if aligned_seqs_indices and verbose:
        print(f"Found {len(aligned_seqs_indices[1])} overlapping residues between the first and second chains.")

    super_imposer = PDB.Superimposer()

    for i in range(1, len(structures)):
        ref_atoms = []
        mod_atoms = []
        for atom_id, (ref_idx, mod_idx) in enumerate(zip(aligned_seqs_indices[0], aligned_seqs_indices[i])):
            ref_coords = coords_list[0][ref_idx]
            mod_coords = coords_list[i][mod_idx]
            shared_atom_names = sorted(set(ref_coords.keys()) & set(mod_coords.keys()))
            for atom_name in shared_atom_names:
                atom = atom_name[0]
                ref_atoms.append(PDB.Atom.Atom("X", ref_coords[atom_name], 1.0, 1.0, " ", "X", atom_id, atom))
                mod_atoms.append(PDB.Atom.Atom("X", mod_coords[atom_name], 1.0, 1.0, " ", "X", atom_id, atom))
        super_imposer.set_atoms(ref_atoms, mod_atoms)
        super_imposer.apply(structures[i].get_atoms())

    aligned_structures_pdb = [get_aligned_structure_as_string(structure) for structure in structures]
    rmsd = super_imposer.rms
    return aligned_structures_pdb, rmsd


def get_aligned_structure_as_string(structure) -> str:
    """Returns the PDB representation of the given structure as a string."""
    io = StringIO()
    pdb_io = PDB.PDBIO()
    pdb_io.set_structure(structure)
    pdb_io.save(io)
    return io.getvalue()


# Taken from AUTOBAN
def pad_line(line):
    """Helper function to pad line to 80 characters in case it is shorter"""
    size_of_line = len(line)
    if size_of_line < 80:
        padding = 80 - size_of_line + 1
        line = line.strip("\n") + " " * padding + "\n"
    return line[:81]  # 80 + newline character


def pdb_to_mmcif_iter(pdb_data: str, structure_id: str, bfactor_to_plddt=False, fractional_plddt=False):
    """
    ADAPTED from pdb-tools pdb_tocif - added section that converts b-factor to pLDDT metric

    Convert a structure in PDB format to mmCIF format.

    This function is a generator.

    Parameters
    ----------
    pdb_data: string with PDB data
    structure_id: entry ID
    bfactor_to_plddt: convert b-factor to alphafold pLDDT metric
    fractional_plddt: multiply pLDDT by 100 to get 0-100, used for ESMFold PDB which has values in 0-1 range

    Yields
    ------
    str (line-by-line)
        The structure in mmCIF format.
    """
    _pad_line = pad_line

    # The spacing here is just aesthetic purposes when printing the file
    _a = "{:<6s} {:5d} {:2s} {:6s} {:1s} {:3s} {:3s} {:1s} {:5s} {:1s} "
    _a += "{:10.3f} {:10.3f} {:10.3f} {:10.3f} {:10.3f} {:1s} "
    _a += "{:5s} {:3s} {:1s} {:4s} {:1d}\n"

    yield "# Converted to mmCIF by pdb-tools\n"
    yield "#\n"

    yield "data_{}\n".format(structure_id)

    yield "#\n"
    yield "loop_\n"
    yield "_atom_site.group_PDB\n"
    yield "_atom_site.id\n"
    yield "_atom_site.type_symbol\n"
    yield "_atom_site.label_atom_id\n"
    yield "_atom_site.label_alt_id\n"
    yield "_atom_site.label_comp_id\n"
    yield "_atom_site.label_asym_id\n"
    yield "_atom_site.label_entity_id\n"
    yield "_atom_site.label_seq_id\n"
    yield "_atom_site.pdbx_PDB_ins_code\n"
    yield "_atom_site.Cartn_x\n"
    yield "_atom_site.Cartn_y\n"
    yield "_atom_site.Cartn_z\n"
    yield "_atom_site.occupancy\n"
    yield "_atom_site.B_iso_or_equiv\n"
    yield "_atom_site.pdbx_formal_charge\n"
    yield "_atom_site.auth_seq_id\n"
    yield "_atom_site.auth_comp_id\n"
    yield "_atom_site.auth_asym_id\n"
    yield "_atom_site.auth_atom_id\n"
    yield "_atom_site.pdbx_PDB_model_num\n"

    # Coordinate data
    model_no = 1
    serial = 0

    records = ("ATOM", "HETATM")
    models_bfactors = []
    model_bfactors = {}
    for line in pdb_data.splitlines():
        if line.startswith(records):
            line = _pad_line(line)

            record = line[0:6].strip()
            serial += 1

            element = line[76:78].strip()
            if not element:
                element = "?"

            atname = line[12:16].strip()
            atname = atname.replace('"', "'")
            if "'" in atname:
                atname = '"{}"'.format(atname)

            altloc = line[16]
            if altloc == " ":
                altloc = "?"

            resname = line[17:20]
            chainid = line[21]
            if chainid == " ":
                chainid = "?"

            resnum = line[22:26].strip()
            icode = line[26]
            if icode == " ":
                icode = "?"

            x = float(line[30:38])
            y = float(line[38:46])
            z = float(line[46:54])

            occ = float(line[54:60])
            bfac = float(line[60:66])

            charge = line[78:80].strip()
            if charge == "":
                charge = "?"

            entityid = "1"

            if atname == "CA":
                if chainid not in model_bfactors:
                    model_bfactors[chainid] = {}
                model_bfactors[chainid][resnum] = (resname, bfac)

            yield _a.format(
                record,
                serial,
                element,
                atname,
                altloc,
                resname,
                chainid,
                entityid,
                resnum,
                icode,
                x,
                y,
                z,
                occ,
                bfac,
                charge,
                resnum,
                resname,
                chainid,
                atname,
                model_no,
            )

        elif line.startswith("ENDMDL"):
            model_no += 1
            models_bfactors.append(model_bfactors)
            model_bfactors = {}
        else:
            continue

    models_bfactors.append(model_bfactors)

    yield "#\n"  # close block

    if bfactor_to_plddt:
        yield "loop_\n"
        yield "_ma_qa_metric.id\n"
        yield "_ma_qa_metric.mode\n"
        yield "_ma_qa_metric.name\n"
        yield "_ma_qa_metric.software_group_id\n"
        yield "_ma_qa_metric.type\n"
        yield "1 global pLDDT 1 pLDDT\n"
        yield "2 local  pLDDT 1 pLDDT\n"
        yield "#\n"
        for modelid, model_bfactors in enumerate(models_bfactors, start=1):
            mean_bfactor = np.mean(
                [bfactor for chain_bfactors in model_bfactors.values() for resname, bfactor in chain_bfactors.values()]
            )
            if fractional_plddt:
                mean_bfactor *= 100.0
            yield "_ma_qa_metric_global.metric_id    1\n"
            yield f"_ma_qa_metric_global.metric_value {mean_bfactor}\n"
            yield f"_ma_qa_metric_global.model_id     {modelid}\n"
            yield f"_ma_qa_metric_global.ordinal_id   {modelid}\n"
        yield "#\n"
        yield "loop_\n"
        yield "_ma_qa_metric_local.label_asym_id\n"
        yield "_ma_qa_metric_local.label_comp_id\n"
        yield "_ma_qa_metric_local.label_seq_id\n"
        yield "_ma_qa_metric_local.metric_id\n"
        yield "_ma_qa_metric_local.metric_value\n"
        yield "_ma_qa_metric_local.model_id\n"
        yield "_ma_qa_metric_local.ordinal_id\n"
        for modelid, model_bfactors in enumerate(models_bfactors, start=1):
            if not model_bfactors:
                continue
            for chainid, chain_bfactors in model_bfactors.items():
                for resno, (resname, bfactor) in chain_bfactors.items():
                    if fractional_plddt:
                        bfactor *= 100.0
                    yield f"{chainid:2s} {resname} {str(resno):4s} 2 {bfactor:.2f} {str(modelid):2s} {str(resno):4s}\n"
        yield "#\n"


def pdb_to_mmcif(pdb_data: str, structure_id: str, bfactor_to_plddt=False, fractional_plddt=False):
    return "".join(
        pdb_to_mmcif_iter(
            pdb_data,
            structure_id,
            bfactor_to_plddt=bfactor_to_plddt,
            fractional_plddt=fractional_plddt,
        )
    )


class ChainNotFoundError(Exception):
    pass


def get_sequences_from_pdb_str(
    pdb_str: str, chains: list[str] = None, by_residue_number: bool = False
) -> dict[str, str] | dict[str, dict[str, str]]:
    """
    Get the sequence of a structure from the pdb file.

    :param pdb_str: str, PDB file contents as string
    :param chains: list of str, chain IDs to extract sequences from, if None, all chains are extracted
    :param by_residue_number: if True, return a dict with residue numbers (strings) as keys and amino acids as values

    Chain breaks (for example a jump from 123 to 134) are NOT filled with X but ignored.
    """
    assert not pdb_str.endswith(".pdb"), f"Expected PDB contents, got path: {pdb_str}"
    parser = PDBParser(QUIET=True)
    parsed_structure = parser.get_structure("", StringIO(pdb_str))
    model = parsed_structure[0]
    if chains is None:
        chains = [c.id for c in model.get_chains()]
    sequences = {}
    for chain in chains:
        if chain not in model:
            raise ChainNotFoundError(chain)
        sequence = []
        visited_residue_ids = set()
        for residue in model[chain]:
            if residue.id not in visited_residue_ids:
                aa = seq1(residue.get_resname())
                if by_residue_number:
                    sequence.append((str(residue.id[1]) + residue.id[2].strip(), aa))
                else:
                    sequence.append(aa)
            visited_residue_ids.add(residue.id)
        sequences[chain] = dict(sequence) if by_residue_number else "".join(sequence)
    return sequences


def get_remark_header(pdb_path: str) -> tuple[str, list[str]]:
    """
    Get the REMARK header from the PDB file.
    :param pdb_path: str, path to the PDB file
    :return: str, REMARK header
    """
    with open(pdb_path, "r") as f:
        lines = f.readlines()

    all_remark_lines = []
    remark_header = ""
    for line in lines:
        if line.startswith("REMARK   1"):
            all_remark_lines.append(line.strip().removeprefix("REMARK   1 "))
            remark_header += line

    return remark_header, all_remark_lines


REMARK_KEYS = ["Input contig", "Standardized contig", "Chains", "Input hotspots", "Standardized hotspots"]


def get_standardized_remarks_from_pdb_str(pdb_str: str) -> dict[str, str]:
    """Parse "standardized" remarks from the PDB file

    :param pdb_str: PDB file contents as string
    :return: dict, parsed remarks {"Input contig": "A45-46/10-15/A45-46/0 B24-26/5/B24-26/0 C10-20", ...}
    Example: example header to read
        REMARK   1 Input contig: "A45-46/10-15/A45-46/0 B24-26/5/B24-26/0 "
        REMARK   1 Input contig: "C10-20"
        REMARK   1 Standardized contig: "A45-46/13-13/A45-46/0 B24-26/5-5/B24-26/"
        REMARK   1 Standardized contig: "0 C10-20/0"
        REMARK   1 Chains: "A B C"
        REMARK   1 Input hotspots:
        REMARK   1 Standardized hotspots:
    """

    parsed_remarks = {}
    stripped_remark_lines = [
        l.removeprefix("REMARK   1").strip() for l in pdb_str.splitlines() if l.startswith("REMARK   1")
    ]
    for key in REMARK_KEYS:
        # Get lines that begin with this key
        remark_lines = [line.removeprefix(f"{key}:").strip() for line in stripped_remark_lines if line.startswith(key)]
        if not remark_lines:
            # Only parse remarks that exist in the PDB
            continue

        values = []
        for line in remark_lines:
            if not line and key not in ["Input hotspots", "Standardized hotspots"]:
                raise ValueError("Empty JSON value after key")
            if not line and key in ["Input hotspots", "Standardized hotspots"]:
                continue

            values.append(json.loads(line))

        parsed_remarks[key] = "".join(values)

    return parsed_remarks


def trim_pdb_str(pdb_input_string: str, target_chain: str, start_res: int, end_res: int) -> str:
    return filter_pdb_str(pdb_input_string, [f"{target_chain}{start_res}-{end_res}"])


def filter_pdb_str(pdb_input_string: str, segments: list[str], add_ter=False) -> str:
    """Filter a PDB string to only include specified segments.

    :param pdb_input_string: str, input PDB string
    :param segments: list of str, segments to include, e.g. ["A5", "A10-20", "B30-40", "C"]
    :param add_ter: Order ATOMs in PDB based on order in selected_segments, add TER and END in between
    :return: str, filtered PDB string
    """
    # Generate trimmed pdb
    parser = PDB.PDBParser(QUIET=True)
    structure = parser.get_structure("filtered_structure", StringIO(pdb_input_string))
    io = PDB.PDBIO()
    io.set_structure(structure)

    stringIO = StringIO()
    # Save the trimmed structure as a IO string
    if add_ter:
        # Save segments one by one to ensure same order as in segments list
        for segment in segments:
            io.save(stringIO, select=PDBSegmentSelector([segment]))
    else:
        io.save(stringIO, select=PDBSegmentSelector(segments))
    return stringIO.getvalue()


def check_rfdiffusion_input(
    pdb_input_string: str,
    target_chain: str,
    start_res_trimmed_chain: int,
    end_res_trimmed_chain: int,
    max_amide_distance: float = 3.0,
):
    """Make sure that structure does not contain chain breaks
    without a gap in residue numbering (required in refolding step)

    Also make sure that no insertion codes are present.

    Raises ValueError if chain breaks are found.
    """
    parser = PDB.PDBParser(QUIET=True)
    structure = parser.get_structure("structure_to_check", StringIO(pdb_input_string))
    model = structure[0]
    chain = model[target_chain]

    residues_with_insertion_codes = []
    for residue in chain:
        if residue.id[2] and residue.id[2].strip():
            residues_with_insertion_codes.append(f"{residue.id[1]}{residue.id[2]}")
    if residues_with_insertion_codes:
        raise ValueError(
            f"Residues with insertion codes found in chain {target_chain}: "
            f"{', '.join(residues_with_insertion_codes)}. "
            f"Please renumber the structure to remove insertion codes."
        )

    previous_residue = None
    for residue in chain:
        resnum = residue.id[1]
        if resnum < start_res_trimmed_chain or resnum > end_res_trimmed_chain:
            continue

        if previous_residue is not None:
            prev_resnum = previous_residue.id[1]
            if "C" not in previous_residue or "N" not in residue:
                # Ignore positions with missing atoms
                previous_residue = residue
                continue
            prev_C = previous_residue["C"].coord
            curr_N = residue["N"].coord
            distance = np.linalg.norm(curr_N - prev_C)
            if distance > max_amide_distance:
                if resnum == prev_resnum + 1:
                    raise ValueError(
                        f"Chain break detected between residues {prev_resnum} and {resnum} in chain {target_chain} "
                        f"with distance {distance:.2f} Å but no gap in residue numbering. "
                        f"Please renumber the structure to include a gap in residue numbering "
                        f"(use a gap of at least 50 residues to indicate different chains) "
                        f"to ensure that chain breaks are preserved by the refolding step."
                    )
        previous_residue = residue



import os
import argparse
import glob

from Bio.PDB import PDBParser
from Bio.SeqUtils import seq1
from Bio import SeqIO

from esm import pretrained
import torch
import pandas as pd

pdb_parser: PDBParser = PDBParser(QUIET=True)
AMINO_ACIDS = list("ACDEFGHIKLMNPQRSTVWY")


def get_sequences_from_pdb_file(file_path: str, chains: str) -> list[tuple[str, str]]:
    """
    Get pairs of sequences and chain identifiers from the pdb structure file.
    """
    with open(file_path) as f:
        parsed_structure = pdb_parser.get_structure("", f)
        available_chains = [c.get_id() for c in parsed_structure[0].get_chains()]
        seqs_concat = []
        for chain in chains:
            if chain not in available_chains:
                raise ValueError(
                    f"Chain {chain} not found in structure {os.path.basename(file_path)}, available chains: {available_chains}"
                )
            sequence = []
            for residue in parsed_structure[0][chain]:
                if residue.has_id("CA"):
                    sequence.append(seq1(residue.get_resname()))
            seqs_concat.append(("".join(sequence), chain))
        return seqs_concat


def load_model(models_path, model_idx):
    model = f"esm1v_t33_650M_UR90S_{model_idx}.pt"
    model_path = os.path.join(models_path, model)
    model, alphabet = pretrained.load_model_and_alphabet(model_path)
    return (model, alphabet)


def run_wildtype_marginals_predictions(data, model, alphabet):
    model.eval()
    if torch.cuda.is_available():
        model = model.cuda()
        print("Transferred model to GPU")
    device = next(model.parameters()).device
    batch_converter = alphabet.get_batch_converter()
    batch_labels, batch_strs, batch_tokens = batch_converter(data)
    batch_tokens = batch_tokens.to(device)

    with torch.no_grad():
        token_probs = torch.softmax(model(batch_tokens)["logits"], dim=-1)
    return token_probs


def get_results_df(data, token_probs, alphabet):
    results = []
    for (label, sequence), token_prob in zip(data, token_probs):
        struct, chain = label.rsplit("_", 1)
        probs_df = pd.DataFrame(token_prob.cpu()[1 : len(sequence) + 1], columns=alphabet.all_toks)
        # Compute the sum of softmax probabilities for the native sequence
        native_seq_softmax_sum = sum(row[aa] for (_, row), aa in zip(probs_df.iterrows(), sequence))
        results.append((struct, chain, sequence, native_seq_softmax_sum))
    # Group results by structure ID to compute results for each structure (supports multichain logic)
    results_df = (
        pd.DataFrame(results, columns=["id", "chain", "sequence", "native_seq_softmax_sum"])
        .groupby("id")
        .agg({"chain": list, "sequence": list, "native_seq_softmax_sum": "sum"})
        .reset_index()
    )
    results_df["chain"] = results_df["chain"].str.join(",")
    results_df["sequence"] = results_df["sequence"].str.join("")
    # Compute average softmax probability per structure given total sequence length (concatenated sequences of processed chain)
    results_df["native_seq_avg_softmax"] = results_df.apply(
        lambda row: row["native_seq_softmax_sum"] / len(row["sequence"]), axis=1
    )
    results_df.drop(columns=["native_seq_softmax_sum"], inplace=True)
    return results_df


def expand_chains_per_structure(filename: str, seqs_concat: list[tuple[str, str]]) -> dict[str, str]:
    """
    Create a dictionary mapping of structure with chain ID to its corresponding sequences.

    Example:
        >>> expand_chains_per_structure("1abc", [("ACDE", "A"), ("FGHI", "B")])
        {'1abc_A': 'ACDE', '1abc_B': 'FGHI'}
    """
    if not seqs_concat:
        return {}
    return {"_".join([filename, chain]): seq for seq, chain in seqs_concat}


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("input_path", type=str)
    parser.add_argument("output_csv", type=str)
    parser.add_argument("-c", "--chains", type=str, default="A", help="Chain to extract the sequence from")
    parser.add_argument("--esm_models_dir", type=str, help="Path to the ESM models directory")
    options = parser.parse_args()

    chains = [chain.strip() for chain in options.chains.split(",")]

    if os.path.isdir(options.input_path):
        paths = sorted(glob.glob(os.path.join(options.input_path, "*.pdb")))
        print(f"Reading sequences from {len(paths):,} PDBs")
        sequences_by_id = {
            struct_chain_id: sequence
            for path in paths
            for struct_chain_id, sequence in expand_chains_per_structure(
                os.path.splitext(os.path.basename(path))[0], get_sequences_from_pdb_file(path, chains=chains)
            ).items()
        }
    elif options.input_path.endswith((".pdb",)):
        sequences_by_id = {
            struct_chain_id: sequence
            for struct_chain_id, sequence in expand_chains_per_structure(
                os.path.splitext(os.path.basename(options.input_path))[0],
                get_sequences_from_pdb_file(options.input_path, chains=chains),
            ).items()
        }
    elif options.input_path.endswith((".fasta", ".fa")):
        sequences_by_id = {
            record.id.split("|")[0]: str(record.seq) for record in SeqIO.parse(options.input_path, "fasta")
        }
    else:
        raise ValueError("Input must be a directory with PDB files, a PDB file, or a FASTA file")

    models_path = options.esm_models_dir

    data = [(sequence_id, sequence) for sequence_id, sequence in sequences_by_id.items()]

    # TODO: Add window-based prediction for long proteins
    token_probs_by_model = []

    n_models = 5
    for model_idx in range(1, n_models + 1):
        # Load the model
        model, alphabet = load_model(models_path, model_idx)

        # Calculate the probabilities
        token_probs = run_wildtype_marginals_predictions(data, model, alphabet)

        token_probs_by_model.append(token_probs)

    avg_token_probs = sum(token_probs_by_model) / n_models
    result = get_results_df(
        data, avg_token_probs, alphabet
    )  # We use alphabet only for column aa identification and it is the same across models
    result.to_csv(options.output_csv, index=False)



from datetime import datetime
import streamlit as st
from streamlit_timeago import time_ago
from ovo import db, get_scheduler
from ovo.core.database.models import DescriptorJob
from ovo.core.logic.descriptor_logic import update_and_process_descriptors


def refresh_descriptors(design_ids: list[str] | set[str], workflow_names: list[str] = None):
    """Update and process all descriptor jobs results and display a Refresh button and log output if any errors occurred."""

    with st.spinner("Updating descriptor job status..."):
        pending_or_failed_jobs = db.select(
            DescriptorJob,
            project_id=st.session_state.project.id,
            _or=(dict(job_result=None), dict(job_result=False)),
            order_by="-created_date_utc",
        )

        # Only consider jobs with a Workflow that links to one of the design IDs
        pending_or_failed_jobs = [
            j for j in pending_or_failed_jobs if j.workflow and set(design_ids).intersection(j.workflow.design_ids)
        ]

        # Only consider certain workflow classes
        if workflow_names:
            if isinstance(workflow_names, str):
                workflow_names = [workflow_names]
            pending_or_failed_jobs = [
                j for j in pending_or_failed_jobs if j.workflow and j.workflow.name in workflow_names
            ]

        pending_jobs = [j for j in pending_or_failed_jobs if j.job_result is None]
        update_and_process_descriptors(descriptor_jobs=pending_jobs, error_callback=st.error)

        failed_jobs = [j for j in pending_or_failed_jobs if j.job_result is False]

        if num_pending_jobs := sum(j.job_result is None for j in pending_jobs):
            st.info(
                f"Not including results of {num_pending_jobs} ongoing descriptor job"
                + ("s" if num_pending_jobs > 1 else "")
            )
            st.button(":material/refresh: Refresh", key="refresh_details_page")
            time_ago(datetime.now(), prefix="Refreshed", key="refreshed")
            with st.expander("Log output"):
                for job in pending_jobs:
                    st.write(job.workflow.name)
                    if st.button("Show log", key=f"show_log_{job.job_id}"):
                        with st.container(height=400):
                            scheduler = get_scheduler(job.scheduler_key)
                            st.code(scheduler.get_log(job.job_id))

        for job in failed_jobs:
            failed_design_ids = sorted(set(design_ids).intersection(job.workflow.design_ids))
            num_failed_designs = len(failed_design_ids)
            design_ids_str = ", ".join(failed_design_ids[:3]) + ("..." if len(failed_design_ids) > 3 else "")
            with st.expander(
                f"{job.workflow.name} failed for {num_failed_designs:,} {'design' if num_failed_designs == 1 else 'designs'}: {design_ids_str}",
                icon="⚠️",
            ):
                if st.button("Show log", key=f"show_log_{job.job_id}_failed"):
                    with st.container(height=400):
                        scheduler = get_scheduler(job.scheduler_key)
                        st.code(scheduler.get_log(job.job_id))



import os
import argparse
import glob

from typing import List, Tuple

import pandas as pd
import numpy as np
from Bio import PDB
from Bio.PDB import PDBParser
from Bio.PDB.Polypeptide import is_aa

from lambda_function import process_pdb, get_peppatch_set_definitions


COLS_TO_NORMALIZE = [
    # Electrostatic pH 5.5 - positive
    "sum_patch_area_+,electrostatic_5.5",
    "top1_patch_area_+,electrostatic_5.5",
    "top2_patches_sum_area_+,electrostatic_5.5",
    "top3_patches_sum_area_+,electrostatic_5.5",
    "top5_patches_sum_area_+,electrostatic_5.5",
    # Electrostatic pH 5.5 - negative
    "sum_patch_area_-,electrostatic_5.5",
    "top1_patch_area_-,electrostatic_5.5",
    "top2_patches_sum_area_-,electrostatic_5.5",
    "top3_patches_sum_area_-,electrostatic_5.5",
    "top5_patches_sum_area_-,electrostatic_5.5",
    # Electrostatic pH 7.4 - positive
    "sum_patch_area_+,electrostatic_7.4",
    "top1_patch_area_+,electrostatic_7.4",
    "top2_patches_sum_area_+,electrostatic_7.4",
    "top3_patches_sum_area_+,electrostatic_7.4",
    "top5_patches_sum_area_+,electrostatic_7.4",
    # Electrostatic pH 7.4 - negative
    "sum_patch_area_-,electrostatic_7.4",
    "top1_patch_area_-,electrostatic_7.4",
    "top2_patches_sum_area_-,electrostatic_7.4",
    "top3_patches_sum_area_-,electrostatic_7.4",
    "top5_patches_sum_area_-,electrostatic_7.4",
    # Hydrophobic
    "sum_patch_area,hydrophobic",
    "top1_patch_area,hydrophobic",
    "top2_patches_sum_area,hydrophobic",
    "top3_patches_sum_area,hydrophobic",
    "top5_patches_sum_area,hydrophobic",
]


def save_first_model_with_chains(
    pdb_file_path: str, out_pdb_file_path: str, chains_to_keep: List[str] = [], skip_unknown: bool = True
):
    """
    Save the first model of a PDB file if it contains multiple models.
    Also if chains_to_keep is provided, only those chains will be saved.
    If skip_unk is True, chains containing UNK residues will not be saved.

    """

    class _SelectChains(PDB.Select):
        def accept_chain(self, chain):
            if chain.id not in chains_to_keep:
                return False
            # Only keep chains with at least one protein residue
            for residue in chain.get_residues():
                if is_aa(residue.get_resname(), standard=True):
                    return True
            return False

        def accept_residue(self, residue):
            if skip_unknown and residue.get_resname() == "UNK":
                return False
            # Only keep standard amino acids
            return is_aa(residue.get_resname(), standard=True)

    # Check if the PDB file contains multiple models
    parser = PDB.PDBParser(QUIET=True)
    structure = parser.get_structure("structure", pdb_file_path)
    first_model = structure[0]
    io = PDB.PDBIO()
    io.set_structure(first_model)
    io.save(out_pdb_file_path, _SelectChains())
    return


def remove_hetatoms(pdb_file_path: str):
    """
    Remove HETATM lines from a PDB file.
    """
    with open(pdb_file_path, "r") as f:
        structure_data = f.read()

    with open(pdb_file_path, "w") as f_out:
        # Remove HETATM lines
        structure_data = "\n".join([line for line in structure_data.split("\n") if not line.startswith("HETATM")])

        # Write the modified data to a new file
        f_out.write(structure_data)


def remove_overlapping_atoms(structure, pdb_path_chain_subset, threshold=1e-3):
    """
    Remove one atom from each pair that are closer than threshold (in Å).
    Modifies the structure in-place.
    """
    atoms = list(structure.get_atoms())
    coords = np.array([atom.get_coord() for atom in atoms])
    to_remove = set()
    for i in range(len(atoms)):
        if i in to_remove:
            continue
        dists = np.linalg.norm(coords[i + 1 :] - coords[i], axis=1)
        overlapping = np.where(dists < threshold)[0]
        for idx in overlapping:
            to_remove.add(i + 1 + idx)
    # Remove atoms (from parent residue)
    for idx in sorted(to_remove, reverse=True):
        atom = atoms[idx]
        parent = atom.get_parent()
        parent.detach_child(atom.id)
    if len(to_remove) > 0:
        print(f"Removed {len(to_remove)} overlapping atoms.")
        io = PDB.PDBIO()
        io.set_structure(structure)
        io.save(pdb_path_chain_subset)


def preprocess_structure(pdb_file_path: str, out_pdb_file_path: str, chains_to_keep: List[str] = []):
    """
    Preprocess the PDB file by saving the first model and removing HETATM lines.
    """
    save_first_model_with_chains(pdb_file_path, out_pdb_file_path, chains_to_keep)
    remove_hetatoms(out_pdb_file_path)


def normalize_area_columns(
    df: pd.DataFrame, cols_to_normalize: List[str], col_for_norm: str = "total_area"
) -> pd.DataFrame:
    """
    Normalize the area columns in the DataFrame by dividing them by the total area.
    The columns to normalize are defined in the  list.
    """
    for column_name in cols_to_normalize:
        if column_name not in df.columns:
            continue
        try:
            descriptor, method = column_name.split(",")
            normalized_column_name = f"{descriptor}_normalized,{method}"
            df[normalized_column_name] = df[column_name] / df[col_for_norm]
        except ValueError:
            print(f"Could not normalize column {column_name} with {col_for_norm}.")
            continue
    return df


def prepare_chain_group_pdbs(pdb_path, out_dir: str, chains_str: str) -> Tuple[List[str], List[str]]:
    """Prepare PDB files for each chain group specified.

    :param path: Path to PDB file
    :param out_dir: Output directory to save prepared PDB files
    :param chains_str: String specifying chain groups, ";" separated groups, "," separated chains within group.
                   Example: "A,B;C" will analyze chains A and B combined as one entry, and chain C as another entry.
                   Use "FULL" to analyze all available chains as one entry.
                   Use "EACH" to analyze each chain separately.

    :return: List of chain groups (['A', 'B', 'A,B']) and corresponding list of prepared PDB file paths
    """
    with open(pdb_path) as f:
        available_chains = sorted(
            {line[21].strip() for line in f if line.startswith("ATOM") and len(line) > 21 and line[21].strip()}
        )
    print("Available chains:", available_chains)

    chain_groups = []
    for group in chains_str.split(";"):
        if group == "EACH":
            chain_groups.extend(available_chains)
        elif group == "FULL":
            chain_groups.append(",".join(available_chains))
        else:
            if missing_chains := sorted(set(group.split(",")).difference(available_chains)):
                raise ValueError(
                    f"Requested chain {missing_chains} is missing in structure: {pdb_path}, available chains: {available_chains}"
                )
            chain_groups.append(group)

    os.makedirs(out_dir, exist_ok=True)
    paths = []
    for chain_group in chain_groups:
        pdb_path_chain_subset = os.path.join(
            out_dir, f"{os.path.basename(pdb_path).removesuffix('.pdb')}.{chain_group}.pdb"
        )
        preprocess_structure(pdb_path, pdb_path_chain_subset, chain_group.split(","))
        paths.append(pdb_path_chain_subset)

    return chain_groups, paths


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("input_path", type=str)
    parser.add_argument("output_csv", type=str)
    parser.add_argument("-c", "--chains", type=str, default="A", help="Chain to extract the sequence from")
    parser.add_argument("--peppatch_config_path", type=str, default=".", help="Path to peppatch config JSON file")
    options = parser.parse_args()

    paths = sorted(glob.glob(os.path.join(options.input_path, "*.pdb")))
    if not paths:
        raise ValueError(f"No pdb files found in: {options.input_path}")

    config_path = options.peppatch_config_path

    peppatch_set = get_peppatch_set_definitions(config_path)

    results = []
    for i, path in enumerate(paths):
        print(f"Processing {i + 1}/{len(paths)}: {path}")
        basename = os.path.basename(path).removesuffix(".pdb")

        chain_groups, chain_pdb_paths = prepare_chain_group_pdbs(path, "prepared_pdb", chains_str=options.chains)
        for chain_group, pdb_path_chain_subset in zip(chain_groups, chain_pdb_paths):
            # Skip large structures over 5000 residues
            parser = PDBParser(QUIET=True)
            structure = parser.get_structure("structure", pdb_path_chain_subset)
            n_residues = sum(1 for _ in structure.get_residues())
            if n_residues < 1:
                print(f"Skipping {basename} chain group {chain_group}: no standard residues found")
                results.append({"id": basename, "chains": chain_group, "error": f"Skipped as no residues were found"})
                continue
            if n_residues > 5000:
                print(f"Skipping {basename} chain group {chain_group}: {n_residues} residues (over 5000)")
                results.append(
                    {
                        "id": basename,
                        "chains": chain_group,
                        "error": f"Skipped due to {n_residues} residues (over 5000)",
                    }
                )
                continue

            # Run pep-patch
            try:
                # Handle overlapping atoms which cause mdtraj process quit https://github.com/mdtraj/mdtraj/issues/1300
                remove_overlapping_atoms(structure, pdb_path_chain_subset)

                per_residue_patches, global_descriptors = process_pdb(pdb_path_chain_subset, peppatch_set)
                results.append(
                    {
                        "id": basename,
                        "chains": chain_group,
                        **global_descriptors,
                        "error": None,
                    }
                )
            except Exception as e:
                print(f"Error processing {basename} chain group {chain_group}: {e}")
                results.append(
                    {
                        "id": basename,
                        "chains": chain_group,
                        "error": str(e),
                    }
                )

    result = pd.DataFrame(results).set_index("id")

    # Normalize area columns
    result = normalize_area_columns(result, COLS_TO_NORMALIZE, "total_area")
    result.to_csv(options.output_csv, index=True)



import streamlit as st

from ovo.core.database.models_rfdiffusion import RFdiffusionBinderDesignWorkflow

from ovo.core.utils.pdb import trim_pdb_str
from ovo.core.utils.residue_selection import from_contig_to_residues, from_residues_to_chain_breaks
from ovo.app.components.molstar_custom_component import molstar_custom_component, StructureVisualization, ContigsParser


def parameters_trim_structure_component(workflow: RFdiffusionBinderDesignWorkflow):
    st.write("Trim the target chain to save computation time, and to avoid running out of GPU memory")

    if workflow.rfdiffusion_params.hotspots:
        st.write(f"Binding hotspots: {workflow.rfdiffusion_params.hotspots}")
        chains_in_hotspot = set([hotspot[0] for hotspot in workflow.rfdiffusion_params.hotspots.split(",")])
        if len(chains_in_hotspot) == 1 and workflow.target_chain is None:
            # pre-fill target chain if determined by selected hotspot
            workflow.target_chain = list(chains_in_hotspot)[0]

    if workflow.target_chain is None and len(workflow.chains) == 1:
        workflow.target_chain = list(workflow.chains)[0]

    target_chain_col, start_trimmed_col, end_trimmed_col, _ = st.columns([0.15, 0.15, 0.15, 0.45])
    with target_chain_col:
        try:
            selected_idx = list(workflow.chains.keys()).index(workflow.target_chain) if workflow.target_chain else None
        except ValueError:
            # chain ID not present in our list
            selected_idx = None

        target_chain = st.selectbox(
            label="Target chain",
            placeholder=f"{len(workflow.chains)} available chains",
            index=selected_idx,
            key="target_chain",
            options=workflow.chains,
            help="Preview supports only one chain at the moment.",
        )

    if not target_chain:
        workflow.target_chain = None
        workflow.start_res_trimmed_chain = None
        workflow.start_end_trimmed_chain = None
        return

    # List residues from the selected chain
    residues = list(map(int, from_contig_to_residues(workflow.chains.get(target_chain))))
    # Check if selected chain has changed
    target_chain_is_unchanged: bool = workflow.target_chain == target_chain

    # Set the start and end residues based on the selected chain if they are not set or the selected chain has changed
    if not workflow.start_res_trimmed_chain or not target_chain_is_unchanged:
        workflow.start_res_trimmed_chain = residues[0]
    if not workflow.end_res_trimmed_chain or not target_chain_is_unchanged:
        workflow.end_res_trimmed_chain = residues[-1]

    with start_trimmed_col:
        start_res = st.number_input(
            "Start residue",
            min_value=residues[0],
            max_value=residues[-1],
            value=workflow.start_res_trimmed_chain,
            key=f"start_trim_residue_{target_chain}",
            args=(
                target_chain,
                residues,
            ),
        )

        if start_res not in residues:
            start_res = min([res for res in residues if res > start_res], key=lambda res: abs(res - start_res))
            st.warning(
                "Start residue is not in the list of residues (between chain breaks). "
                f"Choosing {target_chain}{start_res} as start residue."
            )

    with end_trimmed_col:
        end_res = st.number_input(
            "End residue",
            min_value=residues[0],
            max_value=residues[-1],
            value=workflow.end_res_trimmed_chain,
            key=f"end_trim_residue_{target_chain}",
            args=(
                target_chain,
                residues,
            ),
        )

        if end_res not in residues:
            end_res = min([res for res in residues if res < end_res], key=lambda res: abs(res - end_res))
            st.warning(
                "End residue is not in the list of residues (between chain breaks). "
                f"Choosing {target_chain}{end_res} as end residue."
            )

    trimmed_residues = [res for res in residues if (res <= end_res) and (res >= start_res)]
    chain_break_segments = from_residues_to_chain_breaks(trimmed_residues)
    if chain_break_segments:
        st.info(
            f"Trimmed region {target_chain}{start_res}-{end_res} contains chain breaks: "
            f"{', '.join(chain_break_segments)}. "
            f"AlphaFold2 should detect these and insert a chain break in the prediction, "
            f"but please verify this in the predicted structure."
        )

    if not end_res > start_res:
        workflow.start_res_trimmed_chain = None
        workflow.end_res_trimmed_chain = None
        st.error("End residue must be greater than start residue.")
        return

    # Check if the hotspots are included in the selected range / chain
    if workflow.rfdiffusion_params.hotspots:
        mock_contig = f"{target_chain}{start_res}-{end_res}"
        check_hotspots(mock_contig, workflow.rfdiffusion_params.hotspots)

    # Check if the start and end residues selection has changed since the last preview
    selection_is_unchanged: bool = (
        (workflow.start_res_trimmed_chain == start_res)
        and (workflow.end_res_trimmed_chain == end_res)
        and target_chain_is_unchanged
    )

    # Update the trimmed chain based on the current selection
    workflow.start_res_trimmed_chain = start_res
    workflow.end_res_trimmed_chain = end_res

    # Update the target_chain based on the current selection
    workflow.target_chain = target_chain

    if not selection_is_unchanged:
        # Reset contig to be recomputed in the preview step
        workflow.rfdiffusion_params.contig = None
        return


def trimmed_structure_visualizer(workflow: RFdiffusionBinderDesignWorkflow, pdb_input_string: str):
    target_chain = workflow.get_target_chain()
    trim_start, trim_end = workflow.get_target_trim_boundary()
    pdb_input_string_trimmed = trim_pdb_str(pdb_input_string, target_chain, trim_start, trim_end)

    parser = ContigsParser()

    left, right = st.columns([2, 1])

    with right:
        colors_and_types = {
            ("uniform", "cartoon"): "Cartoon",
            ("hydrophobicity", "molecular-surface"): "Surface hydrophobicity",
        }
        color, representation_type = st.selectbox(
            "Representation",
            options=list(colors_and_types.keys()),
            format_func=lambda x: colors_and_types[x],
            key="colors_and_types_input",
            label_visibility="collapsed",
        )

    left, right = st.columns([1, 1])

    with left:
        st.write("Input structure")
        molstar_custom_component(
            structures=[
                StructureVisualization(
                    pdb=pdb_input_string,
                    contigs=parser.parse_contigs_str(f"{target_chain}{trim_start}-{trim_end}"),
                    highlighted_selections=workflow.get_selected_segments(),
                    color="uniform",
                    representation_type="cartoon",
                )
            ]
            + (
                [
                    StructureVisualization(
                        pdb=pdb_input_string_trimmed,
                        highlighted_selections=workflow.get_selected_segments(),
                        color=color,
                        representation_type=representation_type,
                    )  # include surface of the trimmed structure when displaying surface
                ]
                if "molecular-surface"
                else []
            ),
            key="inp_trim_structure",
        )

    with right:
        st.write("Trimmed structure")

        molstar_custom_component(
            structures=[
                StructureVisualization(
                    pdb=pdb_input_string_trimmed,
                    highlighted_selections=workflow.get_selected_segments(),
                    color=color,
                    representation_type=representation_type,
                )
            ],
            height="524px",  # to level with other visualization
            key="trimmed_structure",
        )


def check_hotspots(contig: str, hotspots: str):
    if hotspots is None or hotspots == "":
        return

    target_chain = contig[0]
    contig_res = contig[1:].split(" ")[0].removesuffix("/0").split("-")
    start_resnum = int(contig_res[0])
    end_resnum = int(contig_res[-1])

    hotspots_list = hotspots.split(",")
    not_included = []
    for hotspot in hotspots_list:
        hotspot_chain, hotspot_resnum = hotspot[0], int(hotspot[1:])

        if hotspot[0] != target_chain:
            not_included.append(hotspot)
            st.error(f"Hotspot {hotspot} is not in the target chain {target_chain}. ")
            continue

        if hotspot_resnum < start_resnum or hotspot_resnum > end_resnum:
            not_included.append(hotspot)

    if not_included:
        st.error(
            f"Trimmed region {start_resnum}-{end_resnum} does not contain "
            f"some of selected hotspots: {', '.join(not_included)}"
        )



from collections import defaultdict

import streamlit as st

from ovo import db, schedulers
from ovo.app.components.descriptor_job_components import refresh_descriptors
from ovo.app.components.descriptor_table import descriptor_table
from ovo.app.components.download_component import download_descriptor_table
from ovo.app.utils.cached_db import (
    get_cached_pools,
    get_cached_design_ids,
    get_cached_available_descriptors,
)
from ovo.core.database import Pool, Design, DesignJob
from ovo.core.database.descriptors_refolding import REFOLDING_DESCRIPTORS
from ovo.core.database.models_refolding import (
    REFOLDING_TESTS_BY_TYPE,
    RefoldingSupportedDesignWorkflow,
    RefoldingWorkflow,
)
from ovo.core.logic.descriptor_logic import submit_descriptor_workflow, get_wide_descriptor_table


@st.fragment
def refolding_fragment(pool_ids: list[str], design_ids: list[str] | None = None):
    pools = get_cached_pools(pool_ids)

    if design_ids is None:
        # design_ids not explicitly passed, use all accepted designs in the selected pools
        design_ids = get_cached_design_ids(pool_ids=pool_ids, accepted=True)
        if not design_ids:
            st.write(
                "No accepted designs in the selected "
                + ("pools" if len(pool_ids) > 1 else "pool")
                + ". Please mark some designs as accepted in the **Jobs** page."
            )
            return
    else:
        if not design_ids:
            # Empty list explicitly passed to design_ids
            st.write("No designs selected")
            return

    st.header(f"🔁 Refolding | {len(design_ids):,} {'design' if len(design_ids) == 1 else 'designs'}")

    if st.button(
        "Submit Refolding",
        type="primary",
        key="submit_refolding_btn",
    ):
        submit_refolding_dialog(pool_ids, design_ids)

    refresh_descriptors(
        design_ids=design_ids,
        workflow_names=[RefoldingWorkflow.name],
    )

    st.subheader("Results")

    available_descriptor_keys = set(get_cached_available_descriptors(design_ids))
    descriptors = [d for d in REFOLDING_DESCRIPTORS if d.key in available_descriptor_keys]

    if not descriptors:
        st.write("No results yet")
        return

    descriptors_df = get_wide_descriptor_table(
        design_ids=design_ids, descriptor_keys=[d.key for d in descriptors], nested=True
    )
    descriptor_table(design_ids, descriptors_df, descriptors)

    download_descriptor_table(
        f"Refolding_test_{'_'.join(pool_ids)}_{len(design_ids)}_designs",
        design_ids,
        descriptor_keys=[d.key for d in descriptors],
    )


@st.fragment
@st.dialog("Refolding submission", width="medium")
def submit_refolding_dialog(pool_ids: list[str], design_ids: list[str]):
    # Create "empty" element to enable clearing the contents after submitting
    content = st.empty()
    with content.container():
        num_designs = len(design_ids)
        st.write(f"""Submit refolding evaluation for {num_designs:,} {"design" if num_designs == 1 else "designs"}""")

        pools = db.select(Pool, id__in=pool_ids)
        design_workflows_by_pool_id = {
            p.id: db.get(DesignJob, id=p.design_job_id).workflow for p in pools if p.design_job_id
        }
        if any(not p.design_job_id for p in pools):
            st.warning(
                """
                Running refolding on custom PDB files requires:
                
                - Designed chain should be chain A
                - The uploaded structure should be the designed structure, not a predicted structure, 
                  since the goal of the refolding workflow is to compare the design to the prediction.
                - In case of AlphaFold2 evaluation of scaffold designs, the PDB file needs to be standardized in OVO format, 
                  containing a REMARK header with the contig string

                More advanced support for custom files will be implemented in the future.
                """
            )
            if not st.checkbox("Acknowledge"):
                return
            design_type = st.selectbox("Select design type", options=REFOLDING_TESTS_BY_TYPE.keys(), key="design_type")
            if not design_type:
                return
        else:
            design_type = None
            for pool in pools:
                design_workflow = design_workflows_by_pool_id[pool.id]
                if not isinstance(design_workflow, RefoldingSupportedDesignWorkflow):
                    st.error(
                        f"Design workflow '{design_workflow.name}' for pool '{pool.name}' does not support refolding evaluation."
                    )
                    return
                workflow_design_type = design_workflow.get_refolding_design_type()
                if design_type is None:
                    design_type = workflow_design_type
                elif design_type != workflow_design_type:
                    st.error(
                        "Selected pools contain different design types "
                        f"('{design_type}' and '{workflow_design_type}'). Please select pools with the same design type."
                    )
                    return

        st.write(f"Select one or more refolding tests supported for **{design_type}** design:")

        tests = []
        for test, (label, description) in REFOLDING_TESTS_BY_TYPE[design_type].items():
            if st.checkbox(f"**{label}** {description}", key=test):
                tests.append(test)

        scheduler_key = st.selectbox(
            "Scheduler",
            options=list(schedulers.keys()),
            format_func=lambda x: schedulers[x].name,
            key="scheduler_selectbox",
        )

        submit = st.button("Submit", disabled=not tests, type="primary")

    if submit:
        content.empty()
        if not tests:
            st.warning("Please select at least one test.")
            return

        # collect designs by their native PDB path (each workflow job only supports a single native structure)
        st.write("Preparing workflow inputs...")
        groups = defaultdict(list)
        for pool in pools:
            index_by_id = db.select_dict(Design, "id", "contig_index", id__in=design_ids, pool_id=pool.id)
            if not pool.design_job_id:
                groups[None] += list(index_by_id.keys())
                continue
            design_workflow = design_workflows_by_pool_id[pool.id]
            ids_by_index = defaultdict(list)
            for design_id, contig_index in index_by_id.items():
                ids_by_index[contig_index].append(design_id)
            for contig_index, ids in ids_by_index.items():
                native_pdb_path = design_workflow.get_refolding_native_pdb_path(contig_index)
                groups[native_pdb_path] += ids

        for native_pdb_path, group_design_ids in groups.items():
            workflow = RefoldingWorkflow(
                design_type=design_type, tests=tests, design_ids=group_design_ids, native_pdb_path=native_pdb_path
            )
            submit_descriptor_workflow(workflow, scheduler_key, st.session_state.project.id)
        st.rerun()



from typing import Callable

import pandas as pd
import streamlit as st

from ovo import storage, CategoricalResidueDescriptor, ResidueNumberDescriptor, NumericGlobalDescriptor, Descriptor
from ovo.app.components import molstar_custom_component, StructureVisualization, ChainVisualization
from ovo.app.components.descriptor_table import residue_number_descriptor_detail_table
from ovo.app.components.download_component import download_job_designs_component
from ovo.app.components.workflow_visualization_components import visualize_design_sequence
from ovo.app.utils.cached_db import get_cached_design, get_cached_pool
from ovo.app.utils.protein_qc_plots import FLAG_ICONS
from ovo.core.database.descriptors_proteinqc import (
    PEP_PATCH_HYDROPHOBICITY_DESCRIPTORS_BY_KEY,
)
from ovo.core.logic.proteinqc_logic import get_flag_color


def on_dismiss_dialog():
    st.query_params.pop("dialog", None)


@st.dialog("Detail", on_dismiss=on_dismiss_dialog)
def detail_descriptor_dialog(descriptor, descriptor_values: pd.Series, design_id: str = None):
    st.markdown(
        """
    <style>
    div[role="dialog"] {
        width: 80vw !important;
        max-width: 1200px !important;
        height: auto;
    }
    </style>
    """,
        unsafe_allow_html=True,
    )

    st.write(f"**{descriptor.name}**: {descriptor.description}")

    col1, col2 = st.columns([2, 3.5])

    with col1:
        table_data, format_func = detail_table(descriptor, descriptor_values)
        if table_data is None:
            return

    with col2:
        detail_design(descriptor, descriptor_values, table_data, format_func, design_id=design_id)


def detail_button(descriptor, descriptor_values: pd.Series, design_id: str = None):
    if st.button("Show more", key=f"{descriptor.name}_btn"):
        st.query_params["dialog"] = descriptor.key
    if st.query_params.get("dialog") == descriptor.key:
        detail_descriptor_dialog(descriptor, descriptor_values, design_id=design_id)


def detail_table(descriptor, descriptor_values: pd.Series):
    table_data = pd.DataFrame({"value": descriptor_values})
    table_data.index.name = "Design ID"
    styler = None

    if isinstance(descriptor, CategoricalResidueDescriptor):
        value_counts = descriptor_values.value_counts()
        table_data["category_abundance"] = descriptor_values.map(value_counts)
        caption = f"Design sorted by **{descriptor.name}** category abundance"
        table_data = table_data.sort_values(by=["category_abundance", "value"], ascending=[False, True])
        column_config = {
            "design_id": st.column_config.Column("Design ID"),
            "value": st.column_config.TextColumn(descriptor.name),
            "category_abundance": st.column_config.ProgressColumn(
                "Category #",
                help=f"Category abundance among the designs",
                format="%d",
                min_value=0,
                max_value=int(table_data["category_abundance"].max()),
                width="small",
            ),
        }
        format_func = lambda design_id: f"{design_id} | {descriptor.name} = {table_data.value.loc[design_id]}"
    elif isinstance(descriptor, NumericGlobalDescriptor):
        caption = f"Designs sorted by **{descriptor.name}**"
        min_val = descriptor.min_value if descriptor.min_value is not None else float(table_data["value"].min())
        max_val = descriptor.max_value if descriptor.max_value is not None else float(table_data["value"].max())

        # Streamlit shows error if min_val == max_val in ProgressColumn so we adjust slightly
        if min_val == max_val:
            epsilon = 1e-6
            min_val -= epsilon
            max_val += epsilon

        table_data = table_data.sort_values(by="value", ascending=True)
        flags = table_data["value"].apply(get_flag_color, descriptor=descriptor)
        if not flags.isna().all():
            table_data["flag"] = flags
            # style flag column
            colors = {
                "missing": "#eeeeee",  # light grey
                "green": "#edf9ee",  # light green
                "yellow": "#ffffec",  # light yellow
                "orange": "#fff6eb",  # orange
            }
            styler = table_data.style.map(
                lambda value: f"background-color: {colors[value]}",
                subset=["flag"],
            )

        column_config = {
            "design_id": st.column_config.Column("Design ID"),
            "value": st.column_config.ProgressColumn(
                descriptor.name, format="%.2f", min_value=min_val, max_value=max_val, color="#111111"
            ),
            "flag": st.column_config.Column("Flag", width="small"),
        }
        format_func = lambda design_id: f"{design_id} | {descriptor.name} = {table_data.value.loc[design_id]:.2f}"
    elif isinstance(descriptor, ResidueNumberDescriptor):
        table_data, column_config, caption, format_func = residue_number_descriptor_detail_table(
            descriptor, descriptor_values
        )
    else:
        # TODO: Support other descriptor tiles
        st.info("Detail view for this type of descriptor coming soon")
        return None, None

    st.caption(caption)
    st.dataframe(
        styler if styler is not None else table_data,
        column_config=column_config,
        width="stretch",
    )
    return table_data, format_func


def detail_design(
    descriptor: Descriptor,
    descriptor_values: pd.Series,
    table_data: pd.DataFrame,
    format_func: Callable,
    design_id: str = None,
):
    design_id = st.selectbox(
        "Select design:",
        options=table_data.index,
        key="design_id_select",
        format_func=format_func,
        index=0 if design_id is None or design_id not in table_data.index else table_data.index.get_loc(design_id),
    )

    value = descriptor_values.loc[design_id]
    flag = get_flag_color(value, descriptor)
    if flag:
        icon = FLAG_ICONS.get(flag, "")
        if flag == "missing":
            st.markdown(f":grey-background[{icon} Value missing]")
        else:
            # TODO explain thresholds
            #  and maybe compare to percentile in PDB ({descriptor.name} = 123, higher than 95% of PDB)
            st.markdown(f":{flag}-background[{icon} {flag.title()} flag]")

    visualize_design_sequence(design_id)

    design = get_cached_design(design_id)

    left, right = st.columns([1, 2], vertical_alignment="center")

    with left:
        colors_and_types = {
            ("chain-id", "cartoon"): (
                "Cartoon",
                "",
            ),
            ("hydrophobicity", "molecular-surface"): (
                "Surface | Hydrophobicity",
                "Surface colored by hydrophobicity scale: :green-badge[**green** = hydroPHOBIC] :red-badge[**red** = hydroPHILIC]",
            ),
            ("hydrophobicity", "cartoon+ball-and-stick"): (
                "Side chains | Hydrophobicity",
                "Cartoon and side chains colored by hydrophobicity scale: :green-badge[**green** = hydroPHOBIC] :red-badge[**red** = hydroPHILIC]",
            ),
            ("residue-charge", "molecular-surface"): (
                "Surface | Residue charge",
                "Surface colored by residue charge: :blue-badge[**blue** = positive] :red-badge[**red** = negative]",
            ),
            ("residue-charge", "cartoon+ball-and-stick"): (
                "Side chains | Residue charge",
                "Cartoon and side chains colored by residue charge: :blue-badge[**blue** = positive] :red-badge[**red** = negative]",
            ),
        }
        rep_type_index = 0
        if descriptor.key in PEP_PATCH_HYDROPHOBICITY_DESCRIPTORS_BY_KEY:
            rep_type_index = 1
        color, representation_type = st.selectbox(
            "Representation",
            options=list(colors_and_types.keys()),
            format_func=lambda x: colors_and_types[x][0],
            key=f"colors_and_types_input_{descriptor.key}",
            label_visibility="collapsed",
            index=rep_type_index,
        )

    with right:
        st.write(colors_and_types[(color, representation_type)][1])

    left, right = st.columns([1.5, 1])
    with left:
        molstar_custom_component(
            structures=[
                StructureVisualization(
                    pdb=storage.read_file_str(design.structure_path),
                    chains=[
                        ChainVisualization(color=color, representation_type=representation_type, chain_id=chain_id)
                        for chain in design.spec.chains
                        for chain_id in chain.chain_ids
                    ],
                )
            ],
            key="dialog_structure",
            height="300px",
        )
    with right:
        pool = get_cached_pool(design.pool_id)
        st.caption(f"Download {design.id}")
        download_job_designs_component(design_ids=[design_id], pools=[pool], key="single", single_line=False)



from datetime import datetime

from ovo.core.auth import is_running_in_streamlit
from ovo import get_scheduler, db
from ovo.core.database.models import JobMixin


def update_job_status(job: JobMixin, save_on_success: bool = False) -> bool | None:
    """Update job status of JobMixin object (Pool, DescriptorJob, ...) and return the scheduler result.

    Job result is saved, except when save_on_success is False and job succeeded,
    this is used in case we want to process the results downstream and save them atomically.

    Returns scheduler result (None = in progress, True = succeeded, False = Failed)
    """
    try:
        scheduler = get_scheduler(job.scheduler_key)
        job_result: bool | None = scheduler.get_result(job.job_id)

        if job.job_started_date_utc is None and (start_time := scheduler.get_job_start_time(job.job_id)):
            # Save start time if not yet saved
            job.job_started_date_utc = start_time
            db.save(job)

        if job_result is not None:
            # Finished state - Successful or Failed
            job.job_result = job_result
            job.job_finished_date_utc = scheduler.get_job_stop_time(job.job_id)
            if job.job_result == False or save_on_success:
                # Save the job if failed or if we want to save successful jobs
                # (job status is not saved in case we want to process the results and save them atomically)
                db.save(job)

        return job_result
    except Exception as e:
        if is_running_in_streamlit():
            # TODO this is not the cleanest way to handle this but probably the most foolproof
            import streamlit as st

            st.error(f"Failed to update job status: {e}")
            return None
        else:
            raise


def format_job_duration(job: JobMixin):
    """Return job duration formatted as 00d:00h:00m."""
    if not job.job_id:
        return None
    if not job.job_started_date_utc:
        return "Not started"
    elif job.job_finished_date_utc:
        delta = job.job_finished_date_utc - job.job_started_date_utc
    else:
        delta = datetime.utcnow() - job.job_started_date_utc
    # format 00d:00h:00m
    days, remainder = divmod(delta.total_seconds(), 86400)
    hours, remainder = divmod(remainder, 3600)
    minutes, _ = divmod(remainder, 60)
    return f"{int(days):02d}d:{int(hours):02d}h:{int(minutes):02d}m"



from typing import List, Callable

from ovo.core.database import DescriptorJob
from ovo.core.database.models import DescriptorWorkflow, WorkflowTypes, Design, Base
from ovo.core.scheduler.base_scheduler import Scheduler
from dataclasses import dataclass


@dataclass
class ProteinQCTool:
    name: str = None
    tool_key: str = None
    supports_conda: bool = False
    supports_multichain: bool = False


SEQ_COMPOSITION = ProteinQCTool(
    name="Sequence Composition",
    tool_key="seq_composition",
    supports_conda=True,
    supports_multichain=True,
)

ESM_1V = ProteinQCTool(
    name="ESM-1v",
    tool_key="esm_1v",
    supports_conda=True,
    supports_multichain=True,
)

ESM_IF = ProteinQCTool(
    name="ESM-IF",
    tool_key="esm_if",
    supports_conda=True,
    supports_multichain=True,
)

DSSP = ProteinQCTool(
    name="DSSP",
    tool_key="dssp",
    supports_conda=True,
    supports_multichain=True,
)

PEPPATCH = ProteinQCTool(
    name="PEP-Patch",
    tool_key="peppatch",
    supports_conda=False,
    supports_multichain=True,
)

PROTEINSOL = ProteinQCTool(
    name="ProteinSol",
    tool_key="proteinsol",
    supports_conda=True,
    supports_multichain=True,
)

PROTEINQC_TOOLS = [SEQ_COMPOSITION, ESM_1V, ESM_IF, DSSP, PEPPATCH, PROTEINSOL]
PROTEINQC_TOOLS_BY_KEY = {tool.tool_key: tool for tool in PROTEINQC_TOOLS}


@WorkflowTypes.register("ProteinQC workflow")
@dataclass
class ProteinQCWorkflow(DescriptorWorkflow):
    tools: List[str] = None

    def get_pipeline_name(self) -> str:
        return "ovo.proteinqc"

    def prepare_params(self, workdir: str) -> dict:
        # TODO: Submit only if there is not existing running job
        # TODO: Submit only if descriptors are missing
        from ovo.core.logic.descriptor_logic import prepare_proteinqc_params

        return prepare_proteinqc_params(
            workflow=self,
            workdir=workdir,
        )

    def process_results(self, job: DescriptorJob, callback: Callable = None) -> list[Base]:
        from ovo.core.logic.descriptor_logic import read_descriptor_file_values

        descriptor_values = read_descriptor_file_values(
            descriptor_job=job,
            # descriptor key prefix (pipeline|tool_key) -> filename to parse
            filenames={f"proteinqc|{tool}": tool for tool in self.tools},
            # mapping from design.id to ID column in produced file
            design_id_mapping={design_id: design_id for design_id in self.design_ids},
        )
        return descriptor_values + [job]

    def validate(self):
        super().validate()
        if not self.tools:
            raise ValueError("No tools specified for ProteinQC workflow")
        unsupported_multi_chain_tools = []
        for tool_key in self.tools:
            if tool_key not in PROTEINQC_TOOLS_BY_KEY:
                raise ValueError(
                    f"Invalid ProteinQC tool '{tool_key}', available tools: {', '.join(PROTEINQC_TOOLS_BY_KEY.keys())}"
                )
            # Validate tool_key has multcat hain support enabled
            if len(self.chains) > 1 and not PROTEINQC_TOOLS_BY_KEY[tool_key].supports_multichain:
                unsupported_multi_chain_tools.append(tool_key)
        if unsupported_multi_chain_tools:
            raise NotImplementedError(
                f"Some selected tools do not support multiple chains yet: {', '.join(unsupported_multi_chain_tools)}"
            )



from datetime import datetime

import streamlit as st
from streamlit_timeago import time_ago

from ovo import db, get_scheduler, Pool, Design, WorkflowTypes, DesignWorkflow
from ovo.app.components.acceptance_thresholds_components import (
    thresholds_and_histograms_component,
    accept_designs_dialog,
    display_current_thresholds,
    filter_designs_by_thresholds_cached,
)
from ovo.app.components.custom_elements import confirm_download_button
from ovo.app.components.download_component import download_job_designs_component
from ovo.app.components.descriptor_job_components import refresh_descriptors
from ovo.app.components.descriptor_scatterplot import (
    descriptor_scatterplot_pool_details_component,
    descriptor_scatterplot_input_component,
)
from ovo.core.database import DesignJob, UnknownWorkflow
from ovo.core.logic.design_logic import get_workflows_table, process_results
from ovo.core.logic.job_logic import update_job_status
from ovo.app.utils.cached_db import get_cached_pools, get_cached_pool, get_cached_design, get_cached_design_job


@st.fragment
def design_job_detail(pool_ids):
    pools = get_cached_pools(pool_ids)
    pools_by_design_job = {pool.design_job_id: pool for pool in pools if pool.design_job_id}

    design_job_ids = [pool.design_job_id for pool in pools if pool.design_job_id]
    if not design_job_ids:
        st.warning("Selected pools are not associated with a design job")
        return

    design_jobs = db.select(DesignJob, id__in=design_job_ids)

    status_icon = (
        ""
        if all(j.job_result for j in design_jobs)
        else ("⚠️" if any(j.job_result is False for j in design_jobs) else "⏳")
    )
    if len(pools) == 1:
        pool = pools[0]
        st.title(f"{status_icon} {pool.name}")

        left_col, right_col = st.columns([0.85, 0.15], vertical_alignment="center")
        with left_col:
            if pool.description:
                st.write(pool.description)
    else:
        st.title(f"{status_icon} {len(pools)} pools")
        st.markdown("#### " + ", ".join([pool.name for pool in pools]))

    st.subheader("Workflow parameters")
    table = get_workflows_table(jobs=design_jobs)
    table.index = [pools_by_design_job[j.id].id for j in design_jobs]
    st.dataframe(table)

    for job in design_jobs:
        pool = pools_by_design_job[job.id]
        if job.workflow and job.workflow.is_instance(UnknownWorkflow):
            st.warning(f"Pool '{pool.id}' workflow failed to load: {job.workflow.error}")
            with st.expander("Raw data"):
                st.json(job.workflow.data)
        if job.job_result is None:
            with st.spinner(f'Checking status of "{pool.name}"'):
                job_result = update_job_status(job)
            if job_result is True and not pool.processed:
                with st.spinner("Workflow finished, processing designs. This will take a moment..."):
                    progress_bar = st.progress(0)
                    process_results(job, callback=progress_bar.progress, wait=False)
                    progress_bar.empty()
                if job.job_result:
                    st.success(f"Workflow finished: {pool.name}")

        if job.job_result is None:
            scheduler = get_scheduler(job.scheduler_key)
            status_label = scheduler.get_status_label(job.job_id)
            st.info(f"Workflow is {status_label}: {pool.name}")
        elif job.job_result == False:
            st.error(f"Workflow failed: {pool.name}")

        if job.warnings:
            st.write(
                ("1 warning" if len(job.warnings) == 1 else f"{len(job.warnings)} warnings")
                + f" found for pool {pool.name}:"
            )
            for warning in job.warnings:
                st.warning(warning)

        # Show log automatically when in progress or failed
        if job.job_result != True:
            args = {}
            if len(pools) == 1 and job.job_result == False:
                args["expanded"] = True
            with st.expander("Log output", **args):
                with st.container(height=400):
                    scheduler = get_scheduler(job.scheduler_key)
                    st.code(scheduler.get_log(job.job_id))

    num_pools_failed = sum(job.job_result == False for job in design_jobs)
    num_pools_in_progress = sum(job.job_result is None for job in design_jobs)

    if len(pools) == num_pools_failed:
        # All pools failed, nothing to show
        return

    if num_pools_in_progress:
        st.button(":material/refresh: Refresh", key="refresh_details_page")
        time_ago(datetime.now(), prefix="Refreshed", key="refreshed")
        if len(pools) == num_pools_in_progress:
            # All pools are in progress, nothing to show yet
            return
        if num_pools_in_progress:
            st.info(
                f"Not including results of {num_pools_in_progress} ongoing workflow"
                + ("s" if num_pools_in_progress > 1 else "")
            )

    all_design_ids = sorted(db.select_unique_values(Design, "id", pool_id__in=pool_ids))

    refresh_descriptors(
        design_ids=all_design_ids,
    )

    scatterplot_fragment(
        all_design_ids=all_design_ids,
        pools=pools,
        jobs=design_jobs,
    )


@st.fragment
def scatterplot_fragment(all_design_ids: list[str], pools: list[Pool], jobs: list[DesignJob]):
    # Get dictionary of saved thresholds (descriptor key -> (min, max) or None)
    saved_thresholds = {}
    inconsistent_threshold_keys = set()
    workflows = [job.workflow for job in jobs]
    for workflow in workflows:
        if not hasattr(workflow, "acceptance_thresholds") or not workflow.acceptance_thresholds:
            continue
        for descriptor_key, thresholds in workflow.acceptance_thresholds.items():
            if descriptor_key not in saved_thresholds:
                saved_thresholds[descriptor_key] = thresholds
            elif saved_thresholds[descriptor_key] != thresholds:
                st.warning(
                    f"Looking at multiple workflows with different thresholds for {descriptor_key}, showing the first one"
                )
                inconsistent_threshold_keys.add(descriptor_key)

    # Save selected_thresholds into session state when first opening the page
    pool_ids_str = ",".join(p.id for p in pools)
    if "selected_thresholds" not in st.session_state or st.session_state.selected_thresholds_pool_ids != pool_ids_str:
        st.session_state.selected_thresholds = saved_thresholds
        st.session_state.selected_thresholds_pool_ids = pool_ids_str

    show_mode = st.segmented_control(
        "Show",
        options=["Accepted designs", "All designs"],
        key="show_designs",
        default=st.query_params.get("show", "Accepted designs"),
    )

    if show_mode == "Accepted designs":
        accepted_design_ids = db.select_values(Design, "id", id__in=all_design_ids, accepted=True)

        if st.session_state.selected_thresholds != saved_thresholds:
            st.header("Editing thresholds")
        elif accepted_design_ids:
            st.header(
                f"Showing {len(accepted_design_ids)} accepted {'designs' if len(accepted_design_ids) > 1 else 'design'}"
            )
        else:
            st.header("No accepted designs")
            st.warning(
                """
                None of the designs met the acceptance thresholds. 
                You may adjust to less strict thresholds below, or try submitting more designs.
                
                To see all generated designs, select 'All designs' above.
                """
            )

        new_accepted_design_ids, num_accepted_by_descriptor = filter_designs_by_thresholds_cached(
            all_design_ids=all_design_ids,
            thresholds=st.session_state.selected_thresholds,
        )

        if st.session_state.selected_thresholds != saved_thresholds:
            st.write(
                f"Accepted designs based on new thresholds: **{len(new_accepted_design_ids):,} / {len(all_design_ids):,}** ({len(new_accepted_design_ids) / len(all_design_ids):.0%})"
            )
        else:
            st.write(
                f"Accepted designs: **{len(accepted_design_ids):,} / {len(all_design_ids):,}** ({len(accepted_design_ids) / len(all_design_ids):.2%})"
            )

        display_current_thresholds(
            selected_thresholds=st.session_state.selected_thresholds,
            all_design_ids=all_design_ids,
            num_accepted_by_descriptor=num_accepted_by_descriptor,
        )

        if st.session_state.selected_thresholds != saved_thresholds:
            if inconsistent_threshold_keys:
                st.warning(
                    f"Selected pools currently use different thresholds for {' & '.join(inconsistent_threshold_keys)}, confirming will override these with the selected threshold value."
                )
            left, mid, _, _ = st.columns(4)
            if left.button("Confirm thresholds", key="confirm_designs_btn", type="primary", width="stretch"):
                accept_designs_dialog(
                    pools=pools,
                    jobs=jobs,
                    all_design_ids=all_design_ids,
                    new_accepted_design_ids=new_accepted_design_ids,
                    num_accepted_by_descriptor=num_accepted_by_descriptor,
                    selected_thresholds=st.session_state.selected_thresholds,
                )
            if mid.button("Discard changes", width="stretch"):
                st.session_state.selected_thresholds = saved_thresholds
                st.rerun()

        st.subheader("Acceptance thresholds")
        if inconsistent_threshold_keys:
            st.warning(
                f"Looking at multiple workflows with different thresholds for {' & '.join(inconsistent_threshold_keys)}, using the first values."
            )
        new_thresholds = thresholds_and_histograms_component(
            selected_thresholds=st.session_state.selected_thresholds,
            all_design_ids=all_design_ids,
        )

        if st.session_state.selected_thresholds != new_thresholds:
            # User just changed one of the thresholds, update them in session state
            st.session_state.selected_thresholds = new_thresholds
            st.rerun()

        if st.session_state.selected_thresholds != saved_thresholds:
            st.header(
                f"{len(new_accepted_design_ids):,} accepted {'design' if len(new_accepted_design_ids) == 1 else 'designs'} based on new thresholds"
            )
            st.warning(
                "Thresholds have been changed, showing preview of accepted designs based on new thresholds. "
                "Please save using the 'Confirm thresholds' button above to apply changes."
            )
            displayed_design_ids = new_accepted_design_ids
        else:
            displayed_design_ids = accepted_design_ids

    elif show_mode == "All designs":
        st.query_params["show"] = show_mode
        scatterplot_settings = descriptor_scatterplot_input_component(all_design_ids)
        if not scatterplot_settings:
            return

        selected_design_ids, selection_label = descriptor_scatterplot_pool_details_component(
            settings=scatterplot_settings,
            design_ids=all_design_ids,
            highlight_accepted=True,
            selected_thresholds=st.session_state.selected_thresholds,
        )

        if selection_label:
            st.info(f"Selected region: {selection_label}")
            displayed_design_ids = selected_design_ids
            st.header(
                f"Showing {len(displayed_design_ids):,} {'design' if len(displayed_design_ids) == 1 else 'designs'} selected in scatterplot"
            )
        else:
            st.caption(
                ":material/info: Select a region in the scatterplot to display designs that fall within the selected range."
            )
            displayed_design_ids = all_design_ids
            st.header(f"Showing all {len(displayed_design_ids):,} designs")

    else:
        st.error("Select a mode.")
        return

    if not displayed_design_ids:
        st.write("No designs found")
        return

    st.subheader("Download")
    download_job_designs_component(displayed_design_ids, pools)

    st.subheader("Structure visualization")
    workflow_names = set(job.workflow.name for job in jobs if job.workflow)

    visualize_designs_fragment(
        design_ids=displayed_design_ids,
        shared_workflow_name=list(workflow_names)[0] if len(workflow_names) == 1 else None,
    )


@st.fragment
def visualize_designs_fragment(design_ids: list[str], shared_workflow_name: str | None = None):
    if not design_ids:
        st.warning("No designs to show")
        return
    elif len(design_ids) == 1:
        design_id = design_ids[0]
    else:
        indexes = {v: i for i, v in enumerate(design_ids, start=1)}
        design_id = st.selectbox(
            "Select design",
            options=[None] + design_ids,
            format_func=lambda v: "Preview all designs" if v is None else f"{indexes[v]} | {v}",
            index=1,
        )

    if design_id is None:
        WorkflowType = WorkflowTypes.get(shared_workflow_name) if shared_workflow_name else DesignWorkflow

        try:
            WorkflowType.visualize_multiple_designs_structures(design_ids=design_ids)
        except NotImplementedError:
            st.warning(
                f"Visualization of multiple designs not supported by {shared_workflow_name or 'default workflow'}, "
                "please select a single design using the dropdown above."
            )
            return
    else:
        st.subheader(design_id)

        design = get_cached_design(design_id)
        pool = get_cached_pool(design.pool_id)
        if not shared_workflow_name and pool.design_job_id:
            design_job = get_cached_design_job(pool.design_job_id)
            shared_workflow_name = design_job.workflow.name if design_job and design_job.workflow else None

        WorkflowType = WorkflowTypes.get(shared_workflow_name) if shared_workflow_name else DesignWorkflow

        try:
            WorkflowType.visualize_single_design_structures(design_id)
        except Exception as e:
            st.error(f"Error visualizing structure: {e}")

        st.write("### Sequence")

        WorkflowType.visualize_single_design_sequences(design_id)

        st.write(f"### Download {design_id}")

        download_job_designs_component(design_ids=[design_id], pools=[pool], key="single")



import sys
import time
import traceback

from rich.console import Console
import requests
import os
import hashlib
import subprocess
from concurrent.futures import ThreadPoolExecutor
from rich.progress import Progress, TextColumn, BarColumn, DownloadColumn, TimeRemainingColumn
import shutil
import tarfile


console = Console(stderr=True, highlight=False)


class OVOCliError(Exception):
    pass


class OVONotInitializedError(OVOCliError):
    pass


def calculate_file_hash(filename, hash_algorithm="sha256"):
    hash_func = hashlib.new(hash_algorithm)
    with open(filename, "rb") as f:
        while chunk := f.read(8192):  # Read the file in chunks
            hash_func.update(chunk)
    return hash_func.hexdigest()


def download_file(url, file_path, file_hash, progress, task):
    progress.console.print(f"Downloading [gray]{url}[/gray] to [gray]{file_path}[/gray]")
    filename = os.path.basename(file_path)

    if os.path.exists(file_path):
        if os.path.isdir(file_path) or calculate_file_hash(file_path) == file_hash:
            progress.console.print(f"[green]Skipping {filename}, file already exists.[/green]")
            progress.update(task, advance=1)
            return
        progress.console.print(f"[yellow]File {filename} exists but hash mismatch. Redownloading...[/yellow]")

    unpack_path = None
    if url.endswith(".tar") and not filename.endswith(".tar"):
        # treat file_path as directory
        assert "." not in filename, f"Filename should be a directory or end with .tar: {filename}"
        unpack_path = file_path
        file_path = file_path + ".tmp"

    response = requests.get(url, stream=True)
    total_size = int(response.headers.get("content-length", 0))  # Get file size from headers
    with open(file_path, "wb") as file:
        progress.update(task, total=total_size)
        for data in response.iter_content(chunk_size=1024):  # Download in chunks
            file.write(data)
            progress.update(task, advance=len(data))

    # Verify the hash of the downloaded file
    downloaded_hash = calculate_file_hash(file_path)
    if downloaded_hash == file_hash:
        console.print(f"[green]Downloaded {filename} successfully and hash matches![/green]")
    else:
        raise ValueError(
            f"[red]Hash mismatch for {filename}! Expected: {file_hash}, Got: {downloaded_hash} in:[/red] {file_path}"
        )

    if unpack_path is not None:
        # unpack file_path into unpack_path
        tmp_path = unpack_path + "_temp"
        os.makedirs(tmp_path, exist_ok=True)
        with tarfile.open(file_path, "r:*") as tar_ref:
            tar_ref.extractall(tmp_path)
        # move to final path only when successful
        shutil.move(tmp_path, unpack_path)


def download_files(destination_dir: str, file_list: list[tuple[str, str, str]]):
    """Download files in parallel using ThreadPoolExecutor."""

    try:
        # Create progress bar
        with Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            DownloadColumn(),
            TimeRemainingColumn(),
        ) as progress:
            tasks = []

            # Initialize download tasks
            for url, file_name, file_hash in file_list:
                file_path = os.path.join(destination_dir, file_name)
                os.makedirs(os.path.dirname(file_path), exist_ok=True)
                task = progress.add_task(f"Downloading {file_name}", total=1)  # total is set later
                tasks.append((url, file_path, file_hash, progress, task))

            # Use ThreadPoolExecutor to download files concurrently
            with ThreadPoolExecutor(max_workers=8) as executor:
                futures = []
                for url, file_path, file_hash, progress, task in tasks:
                    futures.append(executor.submit(download_file, url, file_path, file_hash, progress, task))

                # Wait for all threads to finish
                for future in futures:
                    future.result()
    except Exception as e:
        time.sleep(3)  # short delay to wait for progress bar to flush
        traceback.print_exc()
        console.print(f"\n[red]Error during download:[/red] {e}\n")
        console.print("\nPlease try again, or download the files manually:\n")
        for url, file_path, file_hash, progress, task in tasks:
            console.print(f'wget {url} -O "{file_path}"')
        console.print("")
        sys.exit(2)


def run_nextflow(args: list[str]):
    if shutil.which("nextflow") is None:
        raise OVOCliError("nextflow binary should be available on PATH or in current environment")

    process = subprocess.run(["nextflow"] + args, capture_output=True, text=True)

    # Check for non-zero return code
    if process.returncode != 0:
        if (
            "Nextflow needs a Java virtual machine to run" in process.stderr
            or "java.lang.UnsupportedClassVersionError" in process.stderr
        ):
            raise OVOCliError(
                "Nextflow requires recent Java to run, please install Java 17+ (Java 21-24 recommended). Common ways to install: conda install -c conda-forge openjdk / apt install openjdk-21-jdk / brew install openjdk@21. When working on HPC, activate a recent Java module (module avail java; module load java/version)"
            )
        else:
            # Raise a custom error if the problem is something else
            error = process.stderr.strip() or process.stdout.strip()
            raise OVOCliError(f"Error running Nextflow: {error}")

    return process.stdout


def init_nextflow():
    return run_nextflow(["-h"])
